import unittest
import binascii
import base58
import enum
from copy import deepcopy
from bitarray import bitarray
from neo3.network import payloads, capabilities
from neo3.core import types, serialization
from neo3.core import cryptography as crypto
from neo3 import settings


class AddrTestCase(unittest.TestCase):
    """
    Tests both Network addresses and the AddrPayload
    """
    @classmethod
    def setUpClass(cls) -> None:
        """
        var capability = new FullNodeCapability(123);

        var ip = IPAddress.Parse("127.0.0.1");
        var network_addr = NetworkAddressWithTime.Create(ip, 0, capability);
        Console.WriteLine($"len: {network_addr.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(network_addr.ToArray()).Replace("-", "")}\'");
        """
        cls.network_addr = payloads.NetworkAddress(address='127.0.0.1:0', timestamp=0,
                                capabilities=[capabilities.FullNodeCapability(start_height=123)])

        cls.addr_payload = payloads.AddrPayload([cls.network_addr])

    def test_network_addr_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 26
        self.assertEqual(expected_len, len(self.network_addr))

    def test_network_addr_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'0000000000000000000000000000FFFF7F00000101107B000000')
        self.assertEqual(expected_data, self.network_addr.to_array())

    def test_network_addr_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_addr = payloads.NetworkAddress.deserialize_from_bytes(self.network_addr.to_array())
        self.assertEqual(self.network_addr.timestamp, deserialized_addr.timestamp)
        self.assertEqual(self.network_addr.address, deserialized_addr.address)
        self.assertEqual(self.network_addr.capabilities[0].start_height, deserialized_addr.capabilities[0].start_height)

    def test_addrpayload_len(self):
        expected_len = 27
        self.assertEqual(expected_len, len(self.addr_payload))

    def test_addrpayload_serialization(self):
        expected_data = binascii.unhexlify(b'010000000000000000000000000000FFFF7F00000101107B000000')
        self.assertEqual(expected_data, self.addr_payload.to_array())

    def test_addrpayload_deserialization(self):
        deserialized_addr_payload = payloads.AddrPayload.deserialize_from_bytes(self.addr_payload.to_array())
        self.assertEqual(len(self.addr_payload.addresses), len(deserialized_addr_payload.addresses))
        self.assertEqual(self.addr_payload.addresses[0].capabilities[0].start_height,
                         deserialized_addr_payload.addresses[0].capabilities[0].start_height)

    def test_addrpayload_deserialization2(self):
        capa = capabilities.ServerCapability(n_type=capabilities.NodeCapabilityType.TCPSERVER, port=123)
        network_addr = payloads.NetworkAddress(address='127.0.0.1:0', timestamp=0, capabilities=[capa])
        addr_payload = payloads.AddrPayload([network_addr])

        deserialized_addr_payload = payloads.AddrPayload.deserialize_from_bytes(addr_payload.to_array())
        self.assertEqual(123, deserialized_addr_payload.addresses[0].capabilities[0].port)

    def test_equality(self):
        addr1 = payloads.NetworkAddress(address="127.0.0.1:0")
        addr2 = payloads.NetworkAddress(address="127.0.0.1:0")
        addr3 = payloads.NetworkAddress(address="127.0.0.2:0")
        self.assertTrue(addr1 == addr2)
        self.assertFalse(addr1 == addr3)
        self.assertFalse(addr1 == object())

    def test_various_dunder(self):
        addr1 = payloads.NetworkAddress(address="127.0.0.1:0")
        self.assertEqual("127.0.0.1:0", str(addr1))
        self.assertNotEqual("127.0.0.1", str(addr1))

        formatted = f"{addr1:>12}"
        self.assertEqual(" 127.0.0.1:0", formatted)

        # an object has to be hashable in order for it to be acceptable as a _dictionary key
        # this tests the __hash__ dunder
        try:
            x = {addr1: 123}
        except Exception as e:
            self.fail(f"Failed to add NetworkAddress to dictionary. Can't hash?: {e}")

        # test __repr__
        self.assertIn("127.0.0.1", repr(addr1))
        self.assertIn("(NEW)", repr(addr1))

    def test_properties(self):
        addr1 = payloads.NetworkAddress(address="127.0.0.1:30333")

        self.assertEqual("127.0.0.1", addr1.ip)
        self.assertEqual(30333, addr1.port)

        self.assertTrue(addr1.is_state_new)
        addr1.set_state_connected()
        self.assertFalse(addr1.is_state_new)

        self.assertTrue(addr1.is_state_connected)
        addr1.set_state_poor()
        self.assertFalse(addr1.is_state_connected)

        self.assertTrue(addr1.is_state_poor)
        addr1.set_state_dead()
        self.assertFalse(addr1.is_state_poor)

        self.assertTrue(addr1.is_state_dead)
        addr1.set_state_new()
        self.assertFalse(addr1.is_state_dead)


class BlockTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
            Transaction tx = new Transaction();
            tx.Nonce = 123;
            tx.SystemFee = 456;
            tx.NetworkFee = 789;
            tx.ValidUntilBlock = 1;
            tx.Attributes = new TransactionAttribute[0];
            tx.Signers = new Signer[] { new Signer() { Account = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7")}};
            tx.Script = new byte[] { 0x1 };
            tx.Witnesses = new Witness[] {new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } }};

            Block b = new Block
            {
                Header = new Header
                {
                    Version = 0,
                    PrevHash = UInt256.Parse("0xf782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                    Timestamp = 123,
                    Nonce = 0,
                    Index = 1,
                    PrimaryIndex = 0,
                    NextConsensus = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654"),
                    Witness = new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } },
                    MerkleRoot = UInt256.Zero
                },
                Transactions = new Transaction[] { tx }
            };
            b.Header.MerkleRoot = MerkleTree.ComputeRoot(b.Transactions.Select(p => p.Hash).ToArray());
            Console.WriteLine($"{b.Size}");
            Console.WriteLine($"{BitConverter.ToString(b.ToArray()).Replace("-", "")}");

            var trimmedBlock = new TrimmedBlock
            {
                Header = b.Header,
                Hashes = b.Transactions.Select(p => p.Hash).ToArray()
            };
            Console.WriteLine($"{trimmedBlock.Size}");
            Console.WriteLine($"{BitConverter.ToString(trimmedBlock.ToArray()).Replace("-", "")}");

        """
        cls.tx = payloads.Transaction(version=0,
                                      nonce=123,
                                      system_fee=456,
                                      network_fee=789,
                                      valid_until_block=1,
                                      attributes=[],
                                      signers=[payloads.Signer(types.UInt160.from_string("e239c7228fa6b46cc0cf43623b2f934301d0b4f7"))],
                                      script=b'\x01',
                                      witnesses=[payloads.Witness(invocation_script=b'', verification_script=b'\x55')])

        cls.header = payloads.Header(version=0,
                                     prev_hash=types.UInt256.from_string("f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                                     timestamp=123,
                                     nonce=0,
                                     index=1,
                                     primary_index=0,
                                     next_consensus=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                     witness=payloads.Witness(invocation_script=b'', verification_script=b'\x55'))
        cls.block = payloads.Block(cls.header,
                                   transactions=[cls.tx])
        cls.block.rebuild_merkle_root()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 168
        self.assertEqual(expected_len, len(self.block))

    def test_equals(self):
        self.assertFalse(None == self.block)
        self.assertFalse(self.block == object())

        # test different hashes
        modified_block = deepcopy(self.block)
        modified_block.header.timestamp = 1
        self.assertFalse(self.block == modified_block)
        self.assertTrue(self.block == self.block)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify("000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F75618ADD6F91FAD691D6A4D430DB27CE5CA607296863E73A23FC3622A415CDD407B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D70100015501007B000000C80100000000000015030000000000000100000001F7B4D00143932F3B6243CFC06CB4A68F22C739E20000010101000155")
        self.assertEqual(expected_data, self.block.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_block = payloads.Block.deserialize_from_bytes(self.block.to_array())
        self.assertEqual(self.block.version, deserialized_block.version)
        self.assertEqual(self.block.prev_hash, deserialized_block.prev_hash)
        self.assertEqual(self.block.timestamp, deserialized_block.timestamp)
        self.assertEqual(self.block.nonce, deserialized_block.nonce)
        self.assertEqual(self.block.index, deserialized_block.index)
        self.assertEqual(self.block.primary_index, deserialized_block.primary_index)
        self.assertEqual(self.block.next_consensus, deserialized_block.next_consensus)
        self.assertEqual(self.block.witness.invocation_script, deserialized_block.witness.invocation_script)
        self.assertEqual(self.block.witness.verification_script, deserialized_block.witness.verification_script)
        self.assertEqual(1, len(deserialized_block.transactions))

    def test_deserialization_no_duplicate_transactions(self):
        # A block should not have duplicate transactions
        block_copy = deepcopy(self.block)
        block_copy.transactions.append(block_copy.transactions[0])
        with self.assertRaises(ValueError) as context:
            payloads.Block.deserialize_from_bytes(block_copy.to_array())
        self.assertIn("Deserialization error - block contains duplicate transaction", str(context.exception))

    def test_deserialization_wrong_merkle_root(self):
        block_copy = deepcopy(self.block)
        block_copy.header.merkle_root = types.UInt256.zero()
        with self.assertRaises(ValueError) as context:
            payloads.Block.deserialize_from_bytes(block_copy.to_array())
        self.assertIn("Deserialization error - merkle root mismatch", str(context.exception))

    def test_inventory_type(self):
        self.assertEqual(payloads.InventoryType.BLOCK, self.block.inventory_type)

    def test_trim(self):
        trimmed_block = self.block.trim()
        self.assertIsInstance(trimmed_block, payloads.TrimmedBlock)
        # captured from C#, see setUpClass() for the capture code
        expected_len = 146
        self.assertEqual(expected_len, len(trimmed_block))

        expected_data = binascii.unhexlify('000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F75618ADD6F91FAD691D6A4D430DB27CE5CA607296863E73A23FC3622A415CDD407B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D701000155015618ADD6F91FAD691D6A4D430DB27CE5CA607296863E73A23FC3622A415CDD40')
        self.assertEqual(expected_data, trimmed_block.to_array())

        deserialized_trimmed_block = payloads.TrimmedBlock.deserialize_from_bytes(trimmed_block.to_array())
        self.assertEqual(trimmed_block.header, deserialized_trimmed_block.header)
        self.assertEqual(trimmed_block.hashes, deserialized_trimmed_block.hashes)
        self.assertEqual(1, len(deserialized_trimmed_block.hashes))


class SignerTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Signer co = new Signer();
        co.Account = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654");
        co.Scopes = WitnessScope.CustomContracts | WitnessScope.CustomGroups;
        co.AllowedContracts = new UInt160[] { UInt160.Parse("5b7074e873973a6ed3708862f219a6fbf4d1c411") };

        ECPoint p = ECPoint.Parse("026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc", ECCurve.Secp256r1);
        co.AllowedGroups = new ECPoint[] { p };

        Console.WriteLine($"{co.Size}");
        Console.WriteLine($"{BitConverter.ToString(co.ToArray()).Replace("-","")}");
        Console.Write(co.ToJson());
        """
        cls.signer = payloads.Signer(types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"))
        cls.signer.scope = payloads.WitnessScope.CUSTOM_CONTRACTS | payloads.WitnessScope.CUSTOM_GROUPS
        cls.signer.allowed_contracts = [types.UInt160.from_string("5b7074e873973a6ed3708862f219a6fbf4d1c411")]
        point = crypto.ECPoint(binascii.unhexlify("026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"), crypto.ECCCurve.SECP256R1, True)
        cls.signer.allowed_groups = [point]

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 76
        self.assertEqual(expected_len, len(self.signer))

    def test_equals(self):
        self.assertFalse(self.signer == None)
        self.assertFalse(self.signer == object())
        signer2 = payloads.Signer._serializable_init()
        self.assertFalse(self.signer == signer2)
        self.assertTrue(self.signer == self.signer)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'54A64CAC1B1073E662933EF3E30B007CD98D67D7300111C4D1F4FBA619F2628870D36E3A9773E874705B01026241E7E26B38BB7154B8AD49458B97FB1C4797443DC921C5CA5774F511A2BBFC')
        self.assertEqual(expected_data, self.signer.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_signer = payloads.Signer.deserialize_from_bytes(self.signer.to_array())
        self.assertEqual(self.signer.account, deserialized_signer.account)
        self.assertEqual(self.signer.scope, deserialized_signer.scope)
        self.assertEqual(self.signer.allowed_contracts, deserialized_signer.allowed_contracts)
        self.assertEqual(self.signer.allowed_groups, deserialized_signer.allowed_groups)

    def test_deserialization_invalid_scope(self):
        data = bytearray(self.signer.to_array())
        # the scope is serialized after the UInt160 `account` attribute
        # we modify that byte to an invalid scope type
        data[20] = 0xFF

        with self.assertRaises(ValueError) as context:
            payloads.Signer.deserialize_from_bytes(data)
        self.assertEqual("Deserialization error - invalid scope. GLOBAL scope not allowed with other scope types", str(context.exception))

    def test_to_json(self):
        # captured from C#, see setUpClass() for the capture code
        expected = {"account":"0xd7678dd97c000be3f33e9362e673101bac4ca654","scopes":"CustomContracts, CustomGroups","allowedcontracts":["0x5b7074e873973a6ed3708862f219a6fbf4d1c411"],"allowedgroups":["026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"]}
        self.assertEqual(expected, self.signer.to_json())

    def test_from_json(self):
        captured = {"account": "0xd7678dd97c000be3f33e9362e673101bac4ca654", "scopes": "CustomContracts, CustomGroups",
                    "allowedcontracts": ["0x5b7074e873973a6ed3708862f219a6fbf4d1c411"],
                    "allowedgroups": ["026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"]}
        signer = payloads.Signer.from_json(captured)
        # relies on test_to_json() to pass
        self.assertEqual(captured, signer.to_json())


class FilterAddTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        FilterAddPayload fa = new FilterAddPayload { Data = new byte[] { 0x1, 0x2 } };
        Console.WriteLine($"{fa.Size}");
        Console.WriteLine($"{BitConverter.ToString(fa.ToArray()).Replace("-", "")}");
        """
        cls.filter = payloads.FilterAddPayload(b'\x01\x02')

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 3
        self.assertEqual(expected_len, len(self.filter))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'020102')
        self.assertEqual(expected_data, self.filter.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_filter = payloads.FilterAddPayload.deserialize_from_bytes(self.filter.to_array())
        self.assertEqual(self.filter.data, deserialized_filter.data)


class FilterLoadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        BloomFilter bf = new BloomFilter(8, 2, 345, new byte[] { 0x1, 0x2 });
        FilterLoadPayload fl = FilterLoadPayload.Create(bf);
        Console.WriteLine($"{fl.Size}");
        Console.WriteLine($"{BitConverter.ToString(fl.ToArray()).Replace("-", "")}");
        """
        bloom = crypto.BloomFilter(8, 2, 345, b'\x01\x02')
        cls.filter = payloads.FilterLoadPayload(bloom)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 7
        self.assertEqual(expected_len, len(self.filter))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'01010259010000')
        self.assertEqual(expected_data, self.filter.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_filter = payloads.FilterLoadPayload.deserialize_from_bytes(self.filter.to_array())
        self.assertEqual(self.filter.filter, deserialized_filter.filter)
        self.assertEqual(self.filter.K, deserialized_filter.K)
        self.assertEqual(self.filter.tweak, deserialized_filter.tweak)

    def test_deserialization_with_invalid_K_value(self):
        invalid_K = 51
        bloom = crypto.BloomFilter(8, invalid_K, 345, b'\x01\x02')
        filter = payloads.FilterLoadPayload(bloom)

        with self.assertRaises(ValueError) as context:
            payloads.FilterLoadPayload.deserialize_from_bytes(filter.to_array())
        self.assertIn("Deserialization error - K exceeds limit of 50", str(context.exception))


class GetBlocksPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        UInt256 hash_start = UInt256.Parse("0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01");
        GetBlocksPayload payload = GetBlocksPayload.Create(hash_start, 2);
        Console.WriteLine($"len: {payload.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.hash_start = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01")
        cls.payload = payloads.GetBlocksPayload(hash_start=cls.hash_start, count=2)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 34
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'01FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A40200')
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = payloads.GetBlocksPayload.deserialize_from_bytes(self.payload.to_array())
        self.assertEqual(self.payload.hash_start, deserialized_payload.hash_start)
        self.assertEqual(2, deserialized_payload.count)


class GetBlockByIndexPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        GetBlockByIndexPayload payload = GetBlockByIndexPayload.Create(1, 2);
        Console.WriteLine($"len: {payload.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.payload = payloads.GetBlockByIndexPayload(index_start=1, count=2)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 6
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'010000000200')
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = payloads.GetBlockByIndexPayload.deserialize_from_bytes(self.payload.to_array())
        self.assertEqual(self.payload.index_start, deserialized_payload.index_start)
        self.assertEqual(2, deserialized_payload.count)

    def test_deserialization_error(self):
        # test exceed max count
        payload = payloads.GetBlockByIndexPayload(index_start=1, count=payloads.HeadersPayload.MAX_HEADERS_COUNT + 1)
        with self.assertRaises(ValueError) as context:
            payloads.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))

        # test 0 count
        payload = payloads.GetBlockByIndexPayload(index_start=1, count=0)
        with self.assertRaises(ValueError) as context:
            payloads.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))

        # test negative count
        payload = payloads.GetBlockByIndexPayload(index_start=1, count=-10)
        with self.assertRaises(ValueError) as context:
            payloads.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))


class HeaderTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Header h = new Header
        {
            Version = 0,
            PrevHash = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"),
            MerkleRoot = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"),
            Timestamp = 0,
            Nonce = 0,
            Index = 123,
            PrimaryIndex = 0,
            NextConsensus = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7"),
            Witness = new Witness
            {
                InvocationScript = new byte[] { 0x1, 0x2},
                VerificationScript = new byte[] { 0x3, 0x4}
            }
        };
        Console.WriteLine(h.Hash);
        Console.WriteLine(h.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(h.ToArray()).Replace("-", "")}\'");
        """
        version = 0
        previous_hash = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01")
        merkleroot = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02")
        timestamp = 0
        nonce = 0
        index = 123
        primary_index = 0
        next_consensus = types.UInt160.from_string("e239c7228fa6b46cc0cf43623b2f934301d0b4f7")
        witness = payloads.Witness(invocation_script=b'\x01\x02', verification_script=b'\x03\x04')

        cls.header = payloads.Header(version, previous_hash, timestamp, nonce, index, primary_index, next_consensus, witness, merkleroot)

    def test_len_and_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 116
        expected_hash = types.UInt256.from_string('f8e49a63506ed1273ec79ec6fc0134803aafc251c7e67ffaf355a443617672c5')
        self.assertEqual(expected_len, len(self.header))
        self.assertEqual(expected_hash, self.header.hash())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E201020102020304')
        self.assertEqual(expected_data, self.header.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_header = payloads.Header.deserialize_from_bytes(self.header.to_array())
        self.assertEqual(self.header.hash(), deserialized_header.hash())
        self.assertEqual(self.header.version, deserialized_header.version)
        self.assertEqual(self.header.prev_hash, deserialized_header.prev_hash)
        self.assertEqual(self.header.merkle_root, deserialized_header.merkle_root)
        self.assertEqual(self.header.timestamp, deserialized_header.timestamp)
        self.assertEqual(self.header.nonce, deserialized_header.nonce)
        self.assertEqual(self.header.index, deserialized_header.index)
        self.assertEqual(self.header.next_consensus, deserialized_header.next_consensus)
        self.assertEqual(self.header.witness.invocation_script, deserialized_header.witness.invocation_script)
        self.assertEqual(self.header.witness.verification_script, deserialized_header.witness.verification_script)

    def test_deserialization_failure1(self):
        # there should be a 1 byte witness object count (fixed to value 1) before the actual witness object.
        # see https://github.com/neo-project/neo/issues/1128
        raw_data = binascii.unhexlify(b'0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E200020102020304')
        deserialized_header = payloads.Header._serializable_init()

        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(raw_data) as br:
                    deserialized_header.deserialize(br)
        self.assertIn("Deserialization error", str(context.exception))
        self.assertIn("Witness object count is 0 must be 1", str(context.exception))

    def test_equals(self):
        self.assertFalse(None == self.header)
        self.assertFalse(self.header == object())

        # test different hashes
        modified_header = deepcopy(self.header)
        modified_header.timestamp = 1
        self.assertFalse(self.header == modified_header)
        self.assertTrue(self.header == self.header)


class HeadersPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Header h1 = new Header
        {
            Version = 0,
            PrevHash = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"),
            MerkleRoot = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"),
            Timestamp = 0,
            Nonce = 0,
            Index = 123,
            PrimaryIndex = 0,
            NextConsensus = UInt160.Parse("8a2b438eaca8b4b2ab6b4524b5a69a45d920c351"),
            Witness = new Witness
            {
                InvocationScript = new byte[] {0x1, 0x2},
                VerificationScript = new byte[] {0x3, 0x4}
            }
        };
        Header h2;
        using (BinaryReader reader = new BinaryReader(new MemoryStream(h1.ToArray())))
        {
            h2 = reader.ReadSerializable<Header>();
        }

        HeadersPayload hp = HeadersPayload.Create(new Header[] { h1, h2 });
        Console.WriteLine(hp.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(hp.ToArray()).Replace("-", "")}\'");
        """
        version = 0
        previous_hash = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01")
        merkleroot = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02")
        timestamp = 0
        nonce = 0
        index = 123
        primary_index = 0
        next_consensus = types.UInt160.from_string("8a2b438eaca8b4b2ab6b4524b5a69a45d920c351")
        witness = payloads.Witness(invocation_script=b'\x01\x02', verification_script=b'\x03\x04')

        h1 = payloads.Header(version, previous_hash, timestamp, nonce, index, primary_index, next_consensus, witness, merkleroot)
        h2 = payloads.Header(version, previous_hash, timestamp, nonce, index, primary_index, next_consensus, witness, merkleroot)
        cls.payload = payloads.HeadersPayload([h1, h2])

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 233
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'020000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B0000000051C320D9459AA6B524456BABB2B4A8AC8E432B8A010201020203040000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B0000000051C320D9459AA6B524456BABB2B4A8AC8E432B8A01020102020304')
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = payloads.HeadersPayload.deserialize_from_bytes(self.payload.to_array())
        self.assertEqual(len(self.payload.headers), len(deserialized_payload.headers))
        self.assertEqual(2, len(self.payload.headers))
        self.assertIsInstance(deserialized_payload.headers[0], payloads.Header)
        self.assertIsInstance(deserialized_payload.headers[1], payloads.Header)


class InventoryPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var u1 = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01");
        var u2 = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02");
        var payload = InvPayload.Create(InventoryType.Block, new UInt256[] { u1, u2 });
        Console.WriteLine(payload.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.u1 = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01")
        cls.u2 = types.UInt256.from_string("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02")
        cls.inv = payloads.InventoryPayload(payloads.InventoryType.BLOCK, [cls.u1, cls.u2])

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 66
        self.assertEqual(expected_len, len(self.inv))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'2C0201FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4')
        self.assertEqual(expected_data, self.inv.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_inv_payload = payloads.InventoryPayload.deserialize_from_bytes(self.inv.to_array())
        self.assertEqual(self.inv.type, deserialized_inv_payload.type)
        self.assertEqual(len(self.inv.hashes), len(deserialized_inv_payload.hashes))
        self.assertEqual(self.inv.hashes[0], deserialized_inv_payload.hashes[0])


class MerkleBlockPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Transaction tx = new Transaction();
        tx.Nonce = 123;
        tx.SystemFee = 456;
        tx.NetworkFee = 789;
        tx.ValidUntilBlock = 1;
        tx.Attributes = new TransactionAttribute[0];
        tx.Signers = new Signer[] { new Signer() { Account = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7")}};
        tx.Script = new byte[] { 0x1 };
        tx.Witnesses = new Witness[0];

        Block b = new Block
        {
            Header = new Header
            {
                Version = 0,
                PrevHash = UInt256.Parse("0xf782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                Timestamp = 123,
                Nonce = 0,
                Index = 1,
                PrimaryIndex = 0,
                NextConsensus = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654"),
                Witness = new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } },
                MerkleRoot = UInt256.Zero
            },
            Transactions = new Transaction[] { tx }
        };
        b.Header.MerkleRoot = MerkleTree.ComputeRoot(b.Transactions.Select(p => p.Hash).ToArray());

        byte[] bytes = { 0x1 };
        BitArray flags = new BitArray(bytes);
        MerkleBlockPayload mbp = MerkleBlockPayload.Create(b, flags);
        Console.WriteLine($"b\'{BitConverter.ToString(mbp.ToArray()).Replace("-", "")}\'");
        """

        cls.tx = payloads.Transaction(version=0,
                                      nonce=123,
                                      system_fee=456,
                                      network_fee=789,
                                      valid_until_block=1,
                                      attributes=[],
                                      signers=[payloads.Signer(types.UInt160.from_string("e239c7228fa6b46cc0cf43623b2f934301d0b4f7"))],
                                      script=b'\x01',
                                      witnesses=[payloads.Witness(invocation_script=b'', verification_script=b'\x55')])

        cls.header = payloads.Header(version=0,
                                     prev_hash=types.UInt256.from_string("f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                                     timestamp=123,
                                     nonce=0,
                                     index=1,
                                     primary_index=0,
                                     next_consensus=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                     witness=payloads.Witness(invocation_script=b'', verification_script=b'\x55'))
        cls.block = payloads.Block(cls.header,
                                   transactions=[cls.tx])
        cls.block.rebuild_merkle_root()
        flags = bitarray()
        flags.frombytes(b'\x01')
        cls.merkle_payload = payloads.MerkleBlockPayload(cls.block, flags)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 152
        self.assertEqual(expected_len, len(self.merkle_payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F75618ADD6F91FAD691D6A4D430DB27CE5CA607296863E73A23FC3622A415CDD407B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D70100015501015618ADD6F91FAD691D6A4D430DB27CE5CA607296863E73A23FC3622A415CDD400101')
        self.assertEqual(expected_data, self.merkle_payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_merkle_payload = payloads.MerkleBlockPayload.deserialize_from_bytes(self.merkle_payload.to_array())
        # not testing all properties again. It re-uses the same block as created in the Block test case
        # only testing new properties
        self.assertEqual(self.merkle_payload.tx_count, deserialized_merkle_payload.tx_count)
        self.assertEqual(len(self.merkle_payload.hashes), len(deserialized_merkle_payload.hashes))
        self.assertEqual(self.merkle_payload.hashes[0], deserialized_merkle_payload.hashes[0])
        self.assertEqual(self.merkle_payload.flags, deserialized_merkle_payload.flags)


class PingTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var payload = PingPayload.Create(123, 456);
        payload.Timestamp = 888;
        Console.WriteLine(payload.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.ping = payloads.PingPayload(123)
        cls.ping.nonce = 456
        cls.ping.timestamp = 888

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 12
        self.assertEqual(expected_len, len(self.ping))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'7B00000078030000C8010000')
        self.assertEqual(expected_data, self.ping.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_ping = payloads.PingPayload.deserialize_from_bytes(self.ping.to_array())
        self.assertEqual(self.ping.timestamp, deserialized_ping.timestamp)
        self.assertEqual(self.ping.nonce, deserialized_ping.nonce)
        self.assertEqual(self.ping.current_height, deserialized_ping.current_height)


class TestAttributeTypeEnum(enum.IntEnum):
    TYPE_ZERO = 0x00


class TestTXAttribute(payloads.TransactionAttribute):

    def __init__(self):
        super(TestTXAttribute, self).__init__()

        self.type_ = TestAttributeTypeEnum.TYPE_ZERO
        self.test_member = True

    def __len__(self):
        size_bool = 1
        return super(TestTXAttribute, self).__len__() + size_bool

    def _deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        self.test_member = reader.read_bool()

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_bool(self.test_member)


class TransactionAttributeTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Requires a custom attribute

        class TestTXAttribute : TransactionAttribute
        {
            public override TransactionAttributeType Type { get; }
            public override bool AllowMultiple { get => true; }

            public bool test_member = true;

            public override int Size { get => base.Size + sizeof(bool); }

            protected override void DeserializeWithoutType(BinaryReader reader)
            {
                test_member = reader.ReadBoolean();
            }

            protected override void SerializeWithoutType(BinaryWriter writer)
            {
                writer.Write(test_member);
            }
        }

        TransactionAttribute ta = new TestTXAttribute();
        Console.WriteLine($"{ta.Size}");
        Console.WriteLine($"{BitConverter.ToString(ta.ToArray()).Replace("-", "")}");
        """
        cls.attribute = TestTXAttribute()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 2
        self.assertEqual(expected_len, len(self.attribute))

    def test_equals(self):
        self.assertFalse(self.attribute == None)
        self.assertFalse(self.attribute == object())
        self.assertTrue(self.attribute == TestTXAttribute())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'0001')
        self.assertEqual(expected_data, self.attribute.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_attribute = self.attribute.deserialize_from_bytes(self.attribute.to_array())
        self.assertEqual(self.attribute.type_, deserialized_attribute.type_)
        self.assertEqual(self.attribute.test_member, deserialized_attribute.test_member)

    def test_deserialization_wrong_type(self):
        stream_with_type_one = b'\x01'
        attribute = TestTXAttribute()
        with self.assertRaises(ValueError) as context:
            attribute.deserialize_from_bytes(stream_with_type_one)
        self.assertEqual("Deserialization error - transaction attribute type mismatch", str(context.exception))

    def test_deserialization_from(self):
        stream_with_type_zero = b'\x00\x01'
        with serialization.BinaryReader(stream_with_type_zero) as reader:
            ta = payloads.TransactionAttribute.deserialize_from(reader)
        self.assertIsInstance(ta, TestTXAttribute)

    def test_deserialization_from_failure(self):
        stream_with_invalid_type = b'\xFF'
        with serialization.BinaryReader(stream_with_invalid_type) as reader:
            with self.assertRaises(ValueError) as context:
                payloads.TransactionAttribute.deserialize_from(reader)
            self.assertEqual("Deserialization error - unknown transaction attribute type", str(context.exception))


class TransactionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Transaction tx = new Transaction();
        tx.Nonce = 123;
        tx.SystemFee = 456;
        tx.NetworkFee = 789;
        tx.ValidUntilBlock = 1;
        tx.Attributes = new TransactionAttribute[] { };

        Signer co = new Signer();
        co.Account = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654");
        co.Scopes = WitnessScope.None;
        tx.Signers = new Signer[] { co };

        tx.Script = new byte[] { 0x1, 0x2 };
        tx.Witnesses = new Witness[] { new Witness { InvocationScript=new byte[0], VerificationScript = new byte[] { 0x55 } } };

        Console.WriteLine($"{tx.Size}");
        Console.WriteLine($"{BitConverter.ToString(tx.ToArray()).Replace("-", "")}");
        Console.WriteLine(tx.Hash);
        Console.WriteLine(tx.FeePerByte);
        """
        signer = payloads.Signer(account=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
                                 scope=payloads.WitnessScope.NONE)

        witness = payloads.Witness(invocation_script=b'', verification_script=b'\x55')

        cls.tx = payloads.Transaction(version=0,
                                      nonce=123,
                                      system_fee=456,
                                      network_fee=789,
                                      valid_until_block=1,
                                      attributes=[],
                                      signers=[signer],
                                      script=b'\x01\x02',
                                      witnesses=[witness])

    def tearDown(cls) -> None:
        settings.reset_settings_to_default()

    def test_len_and_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 55
        expected_hash = types.UInt256.from_string('da0343daadf88f95ece657fed6c20e05256d37f0d68757034da1d34f534d2c2c')
        self.assertEqual(expected_len, len(self.tx))
        self.assertEqual(expected_hash, self.tx.hash())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'007B000000C8010000000000001503000000000000010000000154A64CAC1B1073E662933EF3E30B007CD98D67D7000002010201000155')
        self.assertEqual(expected_data, self.tx.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_tx = payloads.Transaction.deserialize_from_bytes(self.tx.to_array())
        self.assertEqual(self.tx.version, deserialized_tx.version)
        self.assertEqual(self.tx.nonce, deserialized_tx.nonce)
        self.assertEqual(self.tx.system_fee, deserialized_tx.system_fee)
        self.assertEqual(self.tx.network_fee, deserialized_tx.network_fee)
        self.assertEqual(self.tx.valid_until_block, deserialized_tx.valid_until_block)
        self.assertEqual(len(self.tx.attributes), len(deserialized_tx.attributes))
        self.assertEqual(len(self.tx.signers), len(deserialized_tx.signers))
        self.assertEqual(self.tx.signers[0].account, deserialized_tx.signers[0].account)
        self.assertEqual(self.tx.signers[0].scope, deserialized_tx.signers[0].scope)
        self.assertEqual(self.tx.script, deserialized_tx.script)
        self.assertEqual(len(self.tx.witnesses), len(deserialized_tx.witnesses))
        self.assertEqual(self.tx.witnesses[0].invocation_script, deserialized_tx.witnesses[0].invocation_script)
        self.assertEqual(self.tx.witnesses[0].verification_script, deserialized_tx.witnesses[0].verification_script)

    def test_deserialization_version_error(self):
        tx = deepcopy(self.tx)
        tx.version = 1
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - invalid version", str(context.exception))

    def test_deserialization_system_fee_error(self):
        tx = deepcopy(self.tx)
        tx.system_fee = -1
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - negative system fee", str(context.exception))

    def test_deserialization_network_fee_error(self):
        tx = deepcopy(self.tx)
        tx.network_fee = -1
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - negative network fee", str(context.exception))

    def test_deserialization_script_length_error(self):
        tx = deepcopy(self.tx)
        tx.script = b''
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - invalid script length 0", str(context.exception))

    def test_deserialization_duplicate_attribute_error(self):
        tx = deepcopy(self.tx)
        tx.attributes = [TestTXAttribute(), TestTXAttribute()]
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - duplicate transaction attribute", str(context.exception))

    def test_deserialization_empty_signers_error(self):
        tx = deepcopy(self.tx)
        tx.signers = []
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - signers can't be empty", str(context.exception))

    def test_deserialization_duplicate_signer(self):
        tx = deepcopy(self.tx)
        tx.signers[0].scope = payloads.WitnessScope.GLOBAL
        tx.signers.append(deepcopy(tx.signers[0]))
        with self.assertRaises(ValueError) as context:
            payloads.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual("Deserialization error - duplicate signer", str(context.exception))

    def test_fee_per_byte(self):
        # captured from C#, see setUpClass() for the capture code
        expected_fee = 14
        self.assertEqual(expected_fee, self.tx.fee_per_byte())

    def test_equals(self):
        self.assertFalse(None == self.tx)
        self.assertFalse(self.tx == object())

        # test different hashes
        modified_tx = deepcopy(self.tx)
        modified_tx.version = 1
        self.assertFalse(self.tx == modified_tx)
        self.assertTrue(self.tx == self.tx)

    def test_inventory_type(self):
        self.assertEqual(payloads.InventoryType.TX, self.tx.inventory_type)

    def test_from_replica(self):
        t = payloads.Transaction._serializable_init()
        t.from_replica(self.tx)
        self.assertEqual(self.tx.version, t.version)
        self.assertEqual(self.tx.nonce, t.nonce)
        self.assertEqual(self.tx.system_fee, t.system_fee)
        self.assertEqual(self.tx.network_fee, t.network_fee)
        self.assertEqual(self.tx.valid_until_block, t.valid_until_block)
        self.assertEqual(len(self.tx.attributes), len(t.attributes))
        self.assertEqual(len(self.tx.signers), len(t.signers))
        self.assertEqual(self.tx.signers[0].account, t.signers[0].account)
        self.assertEqual(self.tx.signers[0].scope, t.signers[0].scope)
        self.assertEqual(self.tx.script, t.script)
        self.assertEqual(len(self.tx.witnesses), len(t.witnesses))
        self.assertEqual(self.tx.witnesses[0].invocation_script, t.witnesses[0].invocation_script)

    def test_special_serialization(self):
        tx_special = deepcopy(self.tx)
        tx_special.block_height = 1
        with serialization.BinaryWriter() as bw:
            tx_special.serialize_special(bw)
            tx_special_bytes = bw.to_array()

        with serialization.BinaryReader(tx_special_bytes) as br:
            tx_from_bytes = payloads.Transaction._serializable_init()
            tx_from_bytes.deserialize_special(br)
        self.assertEqual(tx_special.block_height, tx_from_bytes.block_height)

    def test_sender(self):
        self.assertEqual(self.tx.signers[0].account, self.tx.sender)

        tx = deepcopy(self.tx)
        tx.signers = []
        self.assertEqual(types.UInt160.zero(), tx.sender)

    def test_protocol_magic(self):
        # test proper initialization
        tx = payloads.Transaction._serializable_init()
        # test default
        self.assertEqual(0x4F454E, tx.protocol_magic)
        # test init supplied
        tx = payloads.Transaction(0, 0, 0, 0, 0, protocol_magic=123)
        self.assertEqual(123, tx.protocol_magic)
        # test settings supplied
        settings.network.magic = 456
        tx = payloads.Transaction._serializable_init()
        self.assertEqual(456, tx.protocol_magic)

    def test_scripthashes_for_verifying(self):
        tx = payloads.Transaction._serializable_init()

        account1 = types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654")
        account2 = types.UInt160.zero()
        signer1 = payloads.Signer(account=account1, scope=payloads.WitnessScope.NONE)
        signer2 = payloads.Signer(account=account2, scope=payloads.WitnessScope.NONE)

        tx.signers = [signer1, signer2]

        hashes = tx.get_script_hashes_for_verifying(None)
        self.assertEqual([account1, account2], hashes)

    def test_to_json(self):
        tx = deepcopy(self.tx)
        tx.attributes = [payloads.HighPriorityAttribute()]

        expected = {"hash":"0x61a72c37793ea32b9181105931e7a935eb57c915e9dbeb84dc37068c21df2b20","size":56,"version":0,"nonce":123,"sender":"NTdZAZXWiMSDUTkEFU3nHpT7FEEwwtjwfC","sysfee":"456","netfee":"789","validuntilblock":1,"signers":[{"account":"0xd7678dd97c000be3f33e9362e673101bac4ca654","scopes":"None"}],"attributes":[{"type":"HighPriority"}],"script":"AQI=","witnesses":[{"invocation":"","verification":"VQ=="}]}
        self.assertEqual(expected, tx.to_json())

    def test_from_json(self):
        expected = expected = {"hash":"0x61a72c37793ea32b9181105931e7a935eb57c915e9dbeb84dc37068c21df2b20","size":56,"version":0,"nonce":123,"sender":"NTdZAZXWiMSDUTkEFU3nHpT7FEEwwtjwfC","sysfee":"456","netfee":"789","validuntilblock":1,"signers":[{"account":"0xd7678dd97c000be3f33e9362e673101bac4ca654","scopes":"None"}],"attributes":[{"type":"HighPriority"}],"script":"AQI=","witnesses":[{"invocation":"","verification":"VQ=="}]}
        tx = payloads.Transaction.from_json(expected)
        self.assertEqual(expected, tx.to_json())

class VersionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var capabilities = new List<NodeCapability>
        {
            new FullNodeCapability(123)
        };

        capabilities.Add(new ServerCapability(NodeCapabilityType.TcpServer, (ushort)456));
        NodeCapability[] caps = capabilities.ToArray();
        ISerializable message = VersionPayload.Create(888, "my-user-agent", capabilities.ToArray());
        ((VersionPayload)message).Magic = 769;
        ((VersionPayload)message).Timestamp = 0;
        Console.WriteLine($"b\'{BitConverter.ToString(message.ToArray()).Replace("-", "")}\'");
        """
        settings.network.magic = 769
        capa = [
            capabilities.FullNodeCapability(start_height=123),
            capabilities.ServerCapability(capabilities.NodeCapabilityType.TCPSERVER, port=456)
        ]
        cls.vp = payloads.VersionPayload(nonce=888, user_agent="my-user-agent", capabilities=capa)
        cls.vp.timestamp = 0

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 39
        self.assertEqual(expected_len, len(self.vp))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'010300000000000000000000780300000D6D792D757365722D6167656E7402107B00000001C801')
        self.assertEqual(expected_data, self.vp.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_vp = payloads.VersionPayload.deserialize_from_bytes(self.vp.to_array())
        self.assertEqual(len(self.vp.capabilities), len(deserialized_vp.capabilities))
        self.assertIsInstance(deserialized_vp.capabilities[0], capabilities.FullNodeCapability)
        self.assertEqual(self.vp.capabilities[0].start_height, deserialized_vp.capabilities[0].start_height)
        self.assertIsInstance(deserialized_vp.capabilities[1], capabilities.ServerCapability)
        self.assertEqual(self.vp.capabilities[1].port, deserialized_vp.capabilities[1].port)


class WitnessTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Witness witness = new Witness();
        witness.VerificationScript = new byte[] { 0x01, 0x02 };
        witness.InvocationScript = new byte[] { 0x03, 0x04 };
        Console.WriteLine($"b\'{BitConverter.ToString(witness.ToArray()).Replace("-","")}\'");
        Console.Write(witness.ToJson());
        """
        cls.witness = payloads.Witness(verification_script=b'\x01\x02', invocation_script=b'\x03\x04')

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 6
        self.assertEqual(expected_len, len(self.witness))

    def test_script_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_script_hash = types.UInt160.from_string('9132aaf67ab75c0a604419d920c5cb91e149cc15')
        h = self.witness.script_hash()
        self.assertIsInstance(h, types.UInt160)
        self.assertEqual(expected_script_hash, h)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b'020304020102')
        self.assertEqual(expected_data, self.witness.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_witness = payloads.Witness.deserialize_from_bytes(self.witness.to_array())
        self.assertEqual(self.witness.verification_script, deserialized_witness.verification_script)
        self.assertEqual(self.witness.invocation_script, deserialized_witness.invocation_script)

    def test_to_json(self):
        # captured from C#, see setUpClass() for the capture code
        expected = {"invocation":"AwQ=","verification":"AQI="}
        self.assertEqual(expected, self.witness.to_json())

        expected_empty = {"invocation":"","verification":""}
        empty_witness = payloads.Witness(b'', b'')
        self.assertEqual(expected_empty, empty_witness.to_json())

    def test_from_json(self):
        witness_json = {"invocation": "AwQ=", "verification": "AQI="}
        witness = payloads.Witness.from_json(witness_json)
        self.assertEqual(witness_json, witness.to_json())


class EmptyPayloadTestCase(unittest.TestCase):
    def test_empty(self):
        e = payloads.EmptyPayload()
        self.assertEqual(0, len(e))
        self.assertEqual(0, len(e.to_array()))
        bogus_data = b'\x01\x02\x03'
        self.assertIsInstance(payloads.EmptyPayload.deserialize_from_bytes(bogus_data), payloads.EmptyPayload)



import shutil
import unittest
from sys import stderr
from contextlib import suppress
from tests.storage import storagetest
from neo3.storage import implementations


def setUpModule():
    try:
        options = {'path': './unittest-leveldb'}
        return implementations.LevelDB(options)
    except ModuleNotFoundError as mne:
        print(mne, file=stderr)
        raise


class LevelDBBlocksTestCase(storagetest.AbstractBlockStorageTest):
    def db_factory(self):
        options = {'path': './unittest-leveldb'}
        return implementations.LevelDB(options)

    def tearDown(self) -> None:
        self.db.close()
        with suppress(Exception):
            shutil.rmtree('./unittest-leveldb')


class LevelDBContractsTestCase(storagetest.AbstractContractStorageTest):
    def db_factory(self):
        options = {'path': './unittest-leveldb'}
        return implementations.LevelDB(options)

    def tearDown(self) -> None:
        self.db.close()
        with suppress(Exception):
            shutil.rmtree('./unittest-leveldb')


class LevelDBStoragesTestCase(storagetest.AbstractStorageStorageTest):
    def db_factory(self):
        options = {'path': './unittest-leveldb'}
        return implementations.LevelDB(options)

    def tearDown(self) -> None:
        self.db.close()
        with suppress(Exception):
            shutil.rmtree('./unittest-leveldb')


class LevelDBTransactionsTestCase(storagetest.AbstractTransactionStorageTest):
    def db_factory(self):
        options = {'path': './unittest-leveldb'}
        return implementations.LevelDB(options)

    def tearDown(self) -> None:
        self.db.close()
        with suppress(Exception):
            shutil.rmtree('./unittest-leveldb')


class LevelDBVariousTests(unittest.TestCase):
    def test_plyvel_init_exception(self):
        options = {}
        with self.assertRaises(Exception) as context:
            implementations.LevelDB(options)
        self.assertIn("leveldb exception", str(context.exception))

    def test_no_plyvel_support(self):
        # at this point we know level_db_supported is always true because otherwise it would have failed
        # at `setUpModule` and this test case would never have been called.
        implementations.leveldb.level_db_supported = False
        options = {}
        try:
            with self.assertRaises(Exception) as context:
                implementations.LevelDB(options)
            self.assertIn("plyvel module not found - try 'pip install plyvel", str(context.exception))
        finally:
            implementations.leveldb.level_db_supported = True




from __future__ import annotations
from neo3 import contracts
from neo3.contracts import interop


def register(method: str,
             price: int,
             flags: contracts.CallFlags):
    """
    Register a SYSCALL handler with the Application engine.

    Args:
        method: name of call.
        price: the price of calling the handler.
        flags: ExecutionContext rights needed.
    """
    def inner_func(func):
        interop.InteropService.register(method, func, price, flags)
        return func
    return inner_func



from __future__ import annotations
import hashlib
import abc
import base64
from enum import IntFlag
from neo3.core import serialization, utils, types, cryptography, Size as s, IJson
from neo3.network import payloads
from neo3 import storage
from typing import List, Dict, Any, no_type_check


class Signer(serialization.ISerializable, IJson):
    """
    A class that specifies who can pass CheckWitness() verifications in a smart contract.
    """

    #: Maximum number of allowed_contracts or allowed_groups
    MAX_SUB_ITEMS = 16

    def __init__(self, account: types.UInt160,
                 scope: payloads.WitnessScope = None,
                 allowed_contracts: List[types.UInt160] = None,
                 allowed_groups: List[cryptography.ECPoint] = None):
        #: The TX sender.
        self.account = account
        #: payloads.WitnessScope: The configured validation scope.
        self.scope = scope if scope else payloads.WitnessScope.NONE
        #: List[types.UInt160]: Whitelist of contract script hashes if used with
        #: :const:`~neo3.network.payloads.verification.WitnessScope.CUSTOM_CONTRACTS`.
        self.allowed_contracts = allowed_contracts if allowed_contracts else []
        #: List[cryptography.ECPoint]: Whitelist of public keys if used with
        #: :const:`~neo3.network.payloads.verification.WitnessScope.CUSTOM_GROUPS`.
        self.allowed_groups = allowed_groups if allowed_groups else []

    def __len__(self):
        contracts_size = 0
        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            contracts_size = utils.get_var_size(self.allowed_contracts)

        groups_size = 0
        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            groups_size = utils.get_var_size(self.allowed_groups)

        return s.uint160 + s.uint8 + contracts_size + groups_size

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.account != other.account:
            return False
        return True

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.account)
        writer.write_uint8(self.scope)

        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            writer.write_serializable_list(self.allowed_contracts)

        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            writer.write_serializable_list(self.allowed_groups)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.account = reader.read_serializable(types.UInt160)
        self.scope = payloads.WitnessScope(reader.read_uint8())

        if payloads.WitnessScope.GLOBAL in self.scope and self.scope != payloads.WitnessScope.GLOBAL:
            raise ValueError("Deserialization error - invalid scope. GLOBAL scope not allowed with other scope types")

        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            self.allowed_contracts = reader.read_serializable_list(types.UInt160)

        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            self.allowed_groups = reader.read_serializable_list(cryptography.ECPoint,  # type: ignore
                                                                max=self.MAX_SUB_ITEMS)

    def to_json(self) -> dict:
        """ Convert object into json """
        json: Dict[str, Any] = {
            "account": "0x" + str(self.account),
            "scopes": self.scope.to_csharp_name(),
        }

        if payloads.WitnessScope.CUSTOM_CONTRACTS in self.scope:
            json.update({"allowedcontracts": list(map(lambda a: "0x" + str(a), self.allowed_contracts))})
        if payloads.WitnessScope.CUSTOM_GROUPS in self.scope:
            json.update({"allowedgroups": list(map(lambda g: str(g), self.allowed_groups))})

        return json

    @classmethod
    def from_json(cls, json: dict):
        """ Create object from JSON """
        account = types.UInt160.from_string(json['account'][2:])
        scopes = payloads.WitnessScope.from_chsarp_name(json['scopes'])

        allowed_contracts = []
        if "allowedcontracts" in json:
            for contract in json['allowedcontracts']:
                allowed_contracts.append(types.UInt160.from_string(contract[2:]))

        allowed_groups = []
        if "allowedgroups" in json:
            for group in json['allowedgroups']:
                allowed_groups.append(cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(group)))

        return cls(account, scopes, allowed_contracts, allowed_groups)

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero())


class Witness(serialization.ISerializable, IJson):
    """
    An executable verification script that validates a verifiable object like a transaction.
    """
    _MAX_INVOCATION_SCRIPT = 1024
    _MAX_VERIFICATION_SCRIPT = 1024

    def __init__(self, invocation_script: bytes, verification_script: bytes):
        #: A set of VM instructions to setup the stack for verification.
        self.invocation_script = invocation_script
        #: A set of VM instructions that does the actual verification.
        #: It is expected to set the result stack to a boolean True if validation passed.
        self.verification_script = verification_script
        self._script_hash = None

    def __len__(self):
        return utils.get_var_size(self.invocation_script) + utils.get_var_size(self.verification_script)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_bytes(self.invocation_script)
        writer.write_var_bytes(self.verification_script)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.invocation_script = reader.read_var_bytes(max=self._MAX_INVOCATION_SCRIPT)
        self.verification_script = reader.read_var_bytes(max=self._MAX_VERIFICATION_SCRIPT)

    def script_hash(self) -> types.UInt160:
        """ Get the script hash based on the verification script."""
        intermediate_data = hashlib.sha256(self.verification_script).digest()
        data = hashlib.new('ripemd160', intermediate_data).digest()
        return types.UInt160(data=data)

    def to_json(self) -> dict:
        """ Convert object into json """
        return {
            "invocation": base64.b64encode(self.invocation_script).decode(),
            "verification": base64.b64encode(self.verification_script).decode()
        }

    @classmethod
    def from_json(cls, json: dict):
        """ Create object from JSON """
        return cls(base64.b64decode(json['invocation']), base64.b64decode(json['verification']))

    @classmethod
    def _serializable_init(cls):
        return cls(b'', b'')


class WitnessScope(IntFlag):
    """
    Determine the rules for a smart contract :func:`CheckWitness()` sys call.
    """
    #: No Contract was witnessed. Only sign the transaction.
    NONE = 0x0
    #: Allow the witness if the current calling script hash equals the entry script hash into the virtual machine.
    #: Using this prevents passing :func:`CheckWitness()` in a smart contract called via another smart contract.
    CALLED_BY_ENTRY = 0x01
    #: Allow the witness if called from a smart contract that is whitelisted in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_contracts` attribute.
    CUSTOM_CONTRACTS = 0x10
    #: Allow the witness if any public key is in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_groups` attribute is whitelisted in the contracts
    #: manifest.groups array.
    CUSTOM_GROUPS = 0x20
    #: Allow the witness in all context. Equal to NEO 2.x's default behaviour.
    GLOBAL = 0x80

    @no_type_check
    def to_csharp_name(self) -> str:
        """
        Internal helper to match C# convention
        """
        if self == 0:
            return "None"
        flags = []
        if self.CALLED_BY_ENTRY in self:
            flags.append("CalledByEntry")
        if self.CUSTOM_CONTRACTS in self:
            flags.append("CustomContracts")
        if self.CUSTOM_GROUPS in self:
            flags.append("CustomGroups")
        if self.GLOBAL in self:
            flags.append("Global")
        return ", ".join(flags)

    @classmethod
    def from_chsarp_name(cls, csharp_name):
        """
        Internal helper to parse from C# convention
        """
        c = cls(cls.NONE)
        if "CalledByEntry" in csharp_name:
            c |= c.CALLED_BY_ENTRY
        if "CustomContracts" in csharp_name:
            c |= c.CUSTOM_CONTRACTS
        if "CustomGroups" in csharp_name:
            c |= c.CUSTOM_GROUPS
        if "Global" in csharp_name:
            c |= c.GLOBAL
        return c


class IVerifiable(serialization.ISerializable):
    def __init__(self, *args, **kwargs):
        super(IVerifiable, self).__init__(*args, **kwargs)
        self.witnesses: List[Witness] = []

    @abc.abstractmethod
    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """ """

    @abc.abstractmethod
    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """ """

    @abc.abstractmethod
    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """

    def get_hash_data(self, protocol_magic: int) -> bytes:
        """ Get the unsigned data

        Args:
            protocol_magic: network protocol number (NEO MainNet = 5195086, Testnet = 1951352142, private net = ??)
        """
        with serialization.BinaryWriter() as writer:
            writer.write_uint32(protocol_magic)
            self.serialize_unsigned(writer)
            return writer.to_array()



from __future__ import annotations
import aiohttp
import base64
from dataclasses import dataclass
from typing import List, Dict, Union, Optional, TypedDict, Any, Protocol, Iterator, Tuple
import datetime
from neo3 import contracts
from neo3.network import payloads
from neo3.core import types, cryptography, IJson


@dataclass
class BlockValidator:
    public_key: cryptography.ECPoint
    votes: int
    active: bool

    def __repr__(self):
        return f"{self.__class__.__name__}(public_key={self.public_key}, votes={self.votes}, acitive={self.active})"


@dataclass
class NextBlockValidatorsResponse:
    validators: List[BlockValidator]

    @classmethod
    def from_json(cls, json: dict):
        nvr = cls([])
        for validator in json:
            pk = cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(validator['publickey']))
            votes = int(validator['votes'])
            nvr.validators.append(BlockValidator(pk, votes, validator['active']))
        return nvr


@dataclass
class VersionProtocol:
    addressversion: int
    network: int
    validatorscount: int
    msperblock: int
    maxtraceableblocks: int
    maxtransactionsperblock: int
    maxvaliduntilblockincrement: int
    memorypoolmaxtransactions: int
    initialgasdistribution: int


@dataclass
class GetVersionResponse:
    tcpport: int
    wsport: int
    nonce: int
    useragent: str
    protocol: VersionProtocol

    @classmethod
    def from_json(cls, json: dict):
        json['protocol'] = VersionProtocol(**json['protocol'])
        return cls(**json)


@dataclass
class Peer:
    address: str
    port: int


@dataclass
class GetPeersResponse:
    connected: List[Peer]
    bad: List[Peer]
    unconnected: List[Peer]

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [], [])
        for p in json['connected']:
            c.connected.append(Peer(**p))
        for p in json['bad']:
            c.bad.append(Peer(**p))
        for p in json['unconnected']:
            c.unconnected.append(Peer(**p))
        return c


@dataclass
class Nep17Balance:
    asset_hash: types.UInt160
    amount: int
    last_updated_block: int

    def __repr__(self):
        return f"{self.__class__.__name__}(asset_hash={self.asset_hash}, amount={self.amount}, " \
               f"last_updated_block={self.last_updated_block})"


@dataclass
class Nep17BalancesResponse:
    balances: List[Nep17Balance]
    address: str

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], json['address'])
        for b in json['balance']:
            h = types.UInt160.from_string(b['assethash'][2:])
            a = int(b['amount'])
            c.balances.append(Nep17Balance(h, a, b['lastupdatedblock']))
        return c


@dataclass
class Nep17Transfer:
    time: datetime.datetime
    asset_hash: types.UInt160
    transfer_address: str
    amount: int
    block_index: int
    transfer_notify_index: int
    tx_hash: types.UInt256

    @classmethod
    def from_json(cls, json: dict):
        time = datetime.datetime.fromtimestamp(json['timestamp'] / 1000, datetime.timezone.utc)
        hash_ = types.UInt160.from_string(json['assethash'][2:])
        transfer_addr = json['transferaddress']
        amount = int(json['amount'])
        tx_hash = types.UInt256.from_string(json['txhash'][2:])
        return cls(time, hash_, transfer_addr, amount, json['blockindex'], json['transfernotifyindex'], tx_hash)

    def __repr__(self):
        return f"{self.__class__.__name__}(time={self.time}, asset_hash={self.asset_hash}, " \
               f"transfer_address={self.transfer_address}, amount={self.amount}, block_index={self.block_index}, " \
               f"transfer_notify_index={self.transfer_notify_index}, tx_hash={self.tx_hash})"


@dataclass
class Nep17TransfersResponse:
    sent: List[Nep17Transfer]
    received: List[Nep17Transfer]
    address: str

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [], json['address'])
        for t in json['sent']:
            c.sent.append(Nep17Transfer.from_json(t))
        for t in json['received']:
            c.received.append(Nep17Transfer.from_json(t))
        return c


@dataclass
class MempoolResponse:
    """
    A verified transaction in the memory pool is a transaction which has had:
    - basic structural validation (e.g. max tx size)
    - signature validation
    - state validation
        - block validity expiration check
        - available balance vs network and system fees
        - etc
    """
    verified: List[types.UInt256]
    unverified: List[types.UInt256]

    @classmethod
    def from_json(cls, json: dict):
        c = cls([], [])
        for tx in json['verified']:
            c.verified.append(types.UInt256.from_string(tx[2:]))
        for tx in json['unverified']:
            c.unverified.append(types.UInt256.from_string(tx[2:]))
        return c


@dataclass
class StackItem:
    type: str
    value: Any


class MapStackItem(StackItem):
    def items(self) -> Iterator:
        for pair in self.value:  # type: Tuple[StackItem, StackItem]
            yield pair[0].value, pair[1].value

    def keys(self) -> Iterator:
        for pair in self.value:  # type: Tuple[StackItem, StackItem]
            yield pair[0].value

    def values(self) -> Iterator:
        for pair in self.value:  # type: Tuple[StackItem, StackItem]
            yield pair[1].value

    def __getitem__(self, item: str):
        for pair in self.value:  # type: Tuple[StackItem, StackItem]
            if pair[0].value == item:
                return pair[1].value
        else:
            raise KeyError

    def __iter__(self):
        for pair in self.value:  # type: Tuple[StackItem, StackItem]
            yield pair[0].value


_Item = TypedDict("_Item", {"type": str, "value": Any})


@dataclass
class ExecutionResult:
    state: str
    gas_consumed: int
    exception: Optional[str]
    stack: List[StackItem]

    @staticmethod
    def _parse_stack_item(item: _Item) -> StackItem:
        type_ = item['type']
        if type_ in ("Array", "Struct"):
            list_ = list(map(lambda element: ExecutionResult._parse_stack_item(element), item['value']))
            return StackItem(type_, list_)
        elif type_ in ("Boolean", "Pointer"):
            return StackItem(**item)
        if type_ in ("Buffer", "ByteString"):
            return StackItem(type_, base64.b64decode(item['value']))
        elif type_ == "Integer":
            return StackItem(type_, int(item['value']))
        elif type_ == "Map":
            map_ = []
            for stack_item in item['value']:

                key = ExecutionResult._parse_stack_item(stack_item['key'])
                key_type = stack_item['key']['type']
                if key_type == "ByteString":
                    key.value = key.value.decode()
                else:
                    key.value = str(key.value)
                value = ExecutionResult._parse_stack_item(stack_item['value'])
                map_.append((key, value))
            return MapStackItem(type_, map_)
        elif type_ == "Any":
            return StackItem(type_, None)
        else:
            raise ValueError(f"Unknown stack item type: {type_}")
        assert False, "unreachable"  # just to help mypy

    @classmethod
    def from_json(cls, json: dict):
        gc = int(json['gasconsumed'])
        stack = list(map(lambda item: ExecutionResult._parse_stack_item(item), json['stack']))
        return cls(json['state'], gc, json['exception'], stack)


@dataclass
class ExecutionResultResponse(ExecutionResult):
    script: bytes

    @classmethod
    def from_json(cls, json: dict):
        script = base64.b64decode(json['script'])
        gc = int(json['gasconsumed'])
        stack = list(map(lambda item: ExecutionResult._parse_stack_item(item), json['stack']))
        return cls(json['state'], gc, json['exception'], stack, script)


@dataclass
class Notification:
    contract: types.UInt160
    event_name: str
    state: StackItem

    @classmethod
    def from_json(cls, json: dict):
        c = types.UInt160.from_string(json['contract'][2:])
        e = json['eventname']
        s = ExecutionResult._parse_stack_item(json['state'])
        return cls(c, e, s)

    def __repr__(self):
        return f"{self.__class__.__name__}(contract={str(self.contract)}, " \
               f"event_name={self.event_name}, state={self.state})"


@dataclass
class ApplicationExecution(ExecutionResult):
    trigger: str
    notifications: List[Notification]

    @classmethod
    def from_json(cls, json: dict):
        gc = int(json['gasconsumed'])
        stack = list(map(lambda item: ExecutionResult._parse_stack_item(item), json['stack']))
        state = json['vmstate']
        ex = json['exception']
        notifications = []
        for n in json['notifications']:
            notifications.append(Notification.from_json(n))

        return cls(trigger=json['trigger'], notifications=notifications, state=state, gas_consumed=gc, exception=ex,
                   stack=stack)


@dataclass
class AppliationLogResponse:
    tx_id: types.UInt256
    executions: List[ApplicationExecution]

    @classmethod
    def from_json(cls, json: dict):
        tx_id = types.UInt256.from_string(json['txid'][2:])
        executions = []
        for execution in json['executions']:
            executions.append(ApplicationExecution.from_json(execution))
        return cls(tx_id, executions)

    def __repr__(self):
        return f"{self.__class__.__name__}(tx_id={str(self.tx_id)}, executions={self.executions})"


ContractParameter = Union[bool, int, str, bytes, bytearray, types.UInt160, types.UInt256, cryptography.ECPoint,
                          "ContractParameterArray", "ContractParameterDict"]


class ContractParameterArray(Protocol):
    def insert(self, index: int, value: ContractParameter) -> None: ...

    def __getitem__(self, i: int) -> ContractParameter: ...

    def __setitem__(self, i: int, o: ContractParameter) -> None: ...

    def __delitem__(self, i: int) -> None: ...


class ContractParameterDict(Protocol):
    def __setitem__(self, k: ContractParameter, v: ContractParameter) -> None: ...

    def __delitem__(self, v: ContractParameter) -> None: ...

    def __getitem__(self, k: ContractParameter) -> ContractParameter: ...

    def __iter__(self) -> Iterator[ContractParameter]: ...


class _ContractParameter(IJson):
    def __init__(self, obj: ContractParameter):
        self.value: ContractParameter = ''  # just to help mypy
        if isinstance(obj, bool):
            self.type = contracts.ContractParameterType.BOOLEAN
            self.value = obj
        elif isinstance(obj, int):
            self.type = contracts.ContractParameterType.INTEGER
            self.value = str(obj)
        elif isinstance(obj, str):
            self.type = contracts.ContractParameterType.STRING
            self.value = obj
        elif isinstance(obj, (bytes, bytearray)):
            self.type = contracts.ContractParameterType.BYTEARRAY
            self.value = base64.b64encode(obj).decode()
        elif isinstance(obj, types.UInt160):
            self.type = contracts.ContractParameterType.HASH160
            self.value = f"0x{obj}"
        elif isinstance(obj, types.UInt256):
            self.type = contracts.ContractParameterType.HASH256
            self.value = f"0x{obj}"
        elif isinstance(obj, cryptography.ECPoint):
            self.type = contracts.ContractParameterType.PUBLICKEY
            self.value = obj.to_array().hex()
        elif isinstance(obj, list):
            self.type = contracts.ContractParameterType.ARRAY
            self.value = list(map(lambda element: _ContractParameter(element), obj))
        elif isinstance(obj, dict):
            self.type = contracts.ContractParameterType.MAP
            pairs: List[Dict] = []
            for k, v in obj.items():
                pairs.append({"key": _ContractParameter(k), "value": _ContractParameter(v)})
            # It seems like mypy can't follow that ContractParameter is also
            # a List[Dict[ContractParameter, ContractParameter]]
            self.value = pairs  # type: ignore
        else:
            raise ValueError(f"Unsupported type {type(obj)}")

    @classmethod
    def from_json(cls, json: dict):
        """ Not supported """
        raise NotImplementedError

    def to_json(self) -> dict:
        return {"type": self.type.PascalCase(), "value": self.value}


class RPCClient:
    def __init__(self, url: str, timeout: float = 3.0):
        """
        Args:
            url: host + port
            timeout: total time in seconds a request may take
        """
        self.url = url
        self.timeout = timeout
        self.session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=self.timeout))

    async def _post(self, json: dict):
        """
        Create a POST request with JSON to `self.url` with `self.timeout`

        Raises:
            asyncio.exceptions.TimeoutError
        """
        async with self.session.post(self.url, json=json) as request:
            return await request.json()

    async def close(self):
        """
        Close the client session
        """
        await self.session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.session.closed:
            await self.session.close()


class JsonRpcError(Exception):
    def __init__(self, code: int, message: str, data: str = None):
        self.code = code
        self.message = message
        self.data = "" if data is None else data

    def __str__(self):
        if len(self.data) > 0:
            return f"code={self.code}, message={self.message}, data={self.data}"
        else:
            return f"code={self.code}, message={self.message}"


class NeoRpcClient(RPCClient):
    def __init__(self, host: str, **kwargs):
        super(NeoRpcClient, self).__init__(host, **kwargs)

    async def _do_post(self, method: str, params: List = None, id: int = 0, jsonrpc_version: str = "2.0"):
        params = params if params else []
        json = {'jsonrpc': jsonrpc_version, 'id': id, "method": method, "params": params}
        response = await super(NeoRpcClient, self)._post(json)
        if "error" in response:
            raise JsonRpcError(**response['error'])
        return response['result']

    async def calculate_network_fee(self, transaction: Union[bytes, payloads.Transaction]) -> int:
        """
        Obtain the cost of verifying the transaction and including it in a block (a.k.a network fee).
        """
        if isinstance(transaction, payloads.Transaction):
            transaction = transaction.to_array()
        params = [base64.b64encode(transaction).decode()]
        result = await self._do_post("calculatenetworkfee", params)
        return int(result['networkfee'])

    async def get_application_log(self, tx_hash: Union[types.UInt256, str]) -> AppliationLogResponse:
        """
        Fetch the smart contract event logs for a given transaction.

        Commonly used to verify that a transaction sent via `send_transaction()` was executed succesfully on chain.

        Args:
            tx_hash: the hash of the transaction to query for.
        """
        if isinstance(tx_hash, types.UInt256):
            tx_hash = f"0x{str(tx_hash)}"
        result = await self._do_post("getapplicationlog", [tx_hash])
        return AppliationLogResponse.from_json(result)

    async def get_best_block_hash(self) -> types.UInt256:
        """
        Fetch the hash of the highest block in the chain.
        """
        response = await self._do_post("getbestblockhash")
        return types.UInt256.from_string(response[2:])

    async def get_block(self, index_or_hash: Union[int, types.UInt256]) -> payloads.Block:
        """
        Fetch the block by its index or block hash.
        """
        params: List[Union[int, str]] = []
        if isinstance(index_or_hash, types.UInt256):
            params.append(f"0x{index_or_hash}")
        else:
            params.append(index_or_hash)
        response = await self._do_post("getblock", params)
        return payloads.Block.deserialize_from_bytes(base64.b64decode(response))

    async def get_block_count(self) -> int:
        """
        Fetch the current height of the block chain.
        """
        return await self._do_post("getblockcount")

    async def get_block_hash(self, index: int) -> types.UInt256:
        """
        Fetch the block hash by the block's index.
        """
        response = await self._do_post("getblockhash", [index])
        return types.UInt256.from_string(response[2:])

    async def get_block_header(self, index_or_hash: Union[int, types.UInt256]) -> payloads.Header:
        """
        Fetch the block header by its index or block hash.
        """
        if isinstance(index_or_hash, types.UInt256):
            params = [f"0x{index_or_hash}"]
        else:
            params = [str(index_or_hash)]
        response = await self._do_post("getblockheader", params)
        return payloads.Header.deserialize_from_bytes(base64.b64decode(response))

    async def get_committee(self) -> List[cryptography.ECPoint]:
        """
        Fetch the public keys of the current NEO committee.
        """
        response = await self._do_post("getcommittee")
        keys = []
        for pk in response:
            keys.append(cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(pk)))
        return keys

    async def get_connection_count(self) -> int:
        """
        Fetch the number of peers connected to the node.
        """
        return await self._do_post("getconnectioncount")

    async def get_contract_state(self, contract_hash_or_name: Union[types.UInt160, str]) -> contracts.ContractState:
        """
        Fetch smart contract state information.

        Note:
            Only native contracts can be queried by their name. Name is case-insensitive.
        """
        if isinstance(contract_hash_or_name, types.UInt160):
            params = [f"0x{str(contract_hash_or_name)}"]
        else:
            params = [contract_hash_or_name]
        result = await self._do_post("getcontractstate", params)

        h = types.UInt160.from_string(result['hash'][2:])
        nef = contracts.NEF.from_json(result['nef'])
        manifest = contracts.ContractManifest.from_json(result['manifest'])
        return contracts.ContractState(result['id'], nef, manifest, result['updatecounter'], h)

    async def get_nep17_balances(self, address: str) -> Nep17BalancesResponse:
        """
        Fetch the balance of all NEP17 assets for the specified address.
        """
        result = await self._do_post("getnep17balances", [address])
        return Nep17BalancesResponse.from_json(result)

    async def get_nep17_transfers(self,
                                  address: str,
                                  start_time: Optional[datetime.datetime] = None,
                                  end_time: Optional[datetime.datetime] = None,
                                  ) -> Nep17TransfersResponse:
        """
        Obtain NEP17 transfers for a given address. Defaults to the last 7 days on the server side.

        Args:
            address: account to get transfer for
            start_time: if given the start of the requested range. Must be in UTC and time aware not naïve
            end_time: if given the end of the requested range

        Example:
            # Fetch transfers of the last 14 days
            start = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=14)
            await get_nep17_transfers(<your address>, start)
        """
        params = [address]
        if start_time is not None:
            if start_time.tzinfo is None:
                raise ValueError("start_time is a naïve datetime object which can cause incorrect results. Make it "
                                 "time aware by adding tzinfo. For more information see: "
                                 "https://docs.python.org/3/library/datetime.html#datetime.datetime.tzinfo")
            # C# server side expects timestamp in milliseconds instead of seconds
            t = int(start_time.timestamp() * 1000)
            params.append(str(t))

        if end_time is not None:
            if end_time.tzinfo is None:
                raise ValueError("end_time is a naïve object which can cause incorrect results. Make it time aware by "
                                 "adding tzinfo. For more information see: "
                                 "https://docs.python.org/3/library/datetime.html#datetime.datetime.tzinfo")
            t = int(end_time.timestamp() * 1000)
            params.append(str(t))

        result = await self._do_post("getnep17transfers", params)
        return Nep17TransfersResponse.from_json(result)

    async def get_raw_mempool(self) -> MempoolResponse:
        """
        Return the transaction hashes currently in the memory pool waiting to be added to the next produced block.
        """
        result = await self._do_post("getrawmempool", [True])
        return MempoolResponse.from_json(result)

    async def get_next_blockvalidators(self) -> NextBlockValidatorsResponse:
        """
        Fetch the list of next block validators.
        """
        result = await self._do_post("getnextblockvalidators")
        return NextBlockValidatorsResponse.from_json(result)

    async def get_peers(self) -> GetPeersResponse:
        """
        Fetch peer information
        """
        result = await self._do_post("getpeers")
        return GetPeersResponse.from_json(result)

    async def get_storage(self, script_hash: types.UInt160, key: bytes) -> bytes:
        """
        Fetch a value from a smart contracts storage by its key.

        Args:
            script_hash: contract script hash
            key: the storage key to fetch the data for

        Example:
            # fetch the fee per byte from the Policy native contract
            key_fee_per_byte = b'\x0a'
            await client.get_storage(contracts.PolicyContract().hash, key_fee_per_byte)
        """
        hash_ = f"0x{script_hash}"
        key_encoded = base64.b64encode(key).decode()
        result = await self._do_post("getstorage", params=[hash_, key_encoded])
        return base64.b64decode(result)

    async def get_transaction(self, tx_hash: Union[types.UInt256, str]) -> payloads.Transaction:
        """
        Fetch a transaction by its hash.

        Args:
            tx_hash: as string without "0x" prefix!
        """
        if isinstance(tx_hash, str):
            tx_hash = types.UInt256.from_string(tx_hash)
        result = await self._do_post("getrawtransaction", [f"0x{tx_hash}"])
        return payloads.Transaction.deserialize_from_bytes(base64.b64decode(result))

    async def get_transaction_height(self, tx_hash: Union[types.UInt256, str]) -> int:
        """
        Fetch the height of the block the transaction is included in.

        Args:
            tx_hash: as string without "0x" prefix!
        """
        if isinstance(tx_hash, str):
            tx_hash = types.UInt256.from_string(tx_hash)
        return await self._do_post("gettransactionheight", [f"0x{tx_hash}"])

    async def get_unclaimed_gas(self, address: str) -> int:
        """
        Fetch the amount of unclaimed gas for the given address.

        Args:
            address: a NEO address
        """
        result = await self._do_post("getunclaimedgas", [address])
        return int(result['unclaimed'])

    async def get_version(self) -> GetVersionResponse:
        """
        Fetch the node client version, network protocol properties and network ports.
        """
        return GetVersionResponse.from_json(await self._do_post("getversion"))

    async def invoke_contract_verify(self,
                                     contract_hash: Union[types.UInt160, str],
                                     function_params: Optional[List[ContractParameter]] = None,
                                     signers: Optional[List[payloads.Signer]] = None
                                     ) -> ExecutionResultResponse:
        """
        Invoke the `verify` method on the contract.

        Note:
            Calling smart contracts through this function does not alter the blockchain state.
            The smart contract will be called using the Verification trigger (unlike the `invoke_function` method
            which uses the Application trigger).

        Args:
            contract_hash: the hash of the smart contract to call
            function_params: the arguments required by the smart contract function
            signers: additional signers (e.g. for checkwitness passing)
        """
        if isinstance(contract_hash, str):
            contract_hash = types.UInt160.from_string(contract_hash)
        contract_hash = f"0x{contract_hash}"

        function_params = [] if function_params is None else function_params
        function_params = list(map(lambda fp: _ContractParameter(fp).to_json(), function_params))

        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [contract_hash, function_params, signers]
        result = await self._do_post("invokecontractverify", params)
        return ExecutionResultResponse.from_json(result)

    async def invoke_function(self,
                              contract_hash: Union[types.UInt160, str],
                              name: str,
                              function_params: Optional[List[ContractParameter]] = None,
                              signers: Optional[List[payloads.Signer]] = None
                              ) -> ExecutionResultResponse:
        """
        Invoke a smart contract function.

        Note:
            Calling smart contracts through this function does not alter the blockchain state.
            To alter the blockchain state use the `send_transaction` method instead.

        Args:
            contract_hash: the hash of the smart contract to call
            name: the name of the function to call on the smart contract
            function_params: the arguments required by the smart contract function
            signers: additional signers (e.g. for checkwitness passing)

        Example:
            # check if an account is blocked using the Policy native contract
            policy_contract = "cc5e4edd9f5f8dba8bb65734541df7a1c081c67b"
            account_to_check = types.UInt160.from_string("86df72a6b4ab5335d506294f9ce993722253b6e2")
            signer_account = types.UInt160.from_string("f621168b1fce3a89c33a5f6bcf7e774b4657031c")
            signer = payloads.Signer(signer_account, payloads.WitnessScope.CALLED_BY_ENTRY)
            await client.invoke_function(contract_hash=policy_contract, name="isBlocked",
                                         function_params=[account_to_check], signers=[signer])
        """
        if isinstance(contract_hash, str):
            contract_hash = types.UInt160.from_string(contract_hash)
        contract_hash = f"0x{contract_hash}"

        function_params = [] if function_params is None else function_params
        function_params = list(map(lambda fp: _ContractParameter(fp).to_json(), function_params))

        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [contract_hash, name, function_params, signers]
        result = await self._do_post("invokefunction", params)
        return ExecutionResultResponse.from_json(result)

    async def invoke_script(self,
                            script: bytes,
                            signers: Optional[List[payloads.Signer]] = None
                            ) -> ExecutionResultResponse:
        """
        Executes a script in the virtual machine.

        Note:
            Executing VM scripts through this function does not alter the blockchain state.

        Args:
            script: an array of VM opcodes
            signers: additional signers (e.g. for checkwitness passing)

        Returns:
            The results of executing the script in the VM
        """
        signers = [] if signers is None else signers
        signers = list(map(lambda s: s.to_json(), signers))  # type: ignore

        params = [base64.b64encode(script).decode(), signers]
        result = await self._do_post("invokescript", params)
        return ExecutionResultResponse.from_json(result)

    async def send_transaction(self, tx: Union[payloads.Transaction, bytes]) -> types.UInt256:
        """
        Broadcast a transaction to the network.

        Note:
            uses the `sendrawtransaction` RPC method internally.

        Args:
            tx: either a Transaction object or a serialized Transaction. Must be signed

        Returns:
            a transaction hash if successful.
        """
        if isinstance(tx, payloads.Transaction):
            tx = tx.to_array()
        result = await self._do_post("sendrawtransaction", [base64.b64encode(tx).decode()])
        return types.UInt256.from_string(result['hash'][2:])

    async def send_block(self, block: Union[payloads.Block, bytes]) -> types.UInt256:
        """
        Broadcast a transaction to the network.

        Args:
            block: either a Block object or a serialized Block

        Returns:
            a block hash if successful.
        """
        if isinstance(block, payloads.Block):
            block = block.to_array()
        result = await self._do_post("submitblock", [base64.b64encode(block).decode()])
        return types.UInt256.from_string(result['hash'][2:])

    async def validate_address(self, address: str) -> bool:
        """
        Verify if the given address is valid for the network the node is running on.

        Args:
            address: a NEO address
        """
        result = await self._do_post("validateaddress", [address])
        return result['isvalid']

    async def print_contract_methods(self, contract_hash_or_name: Union[types.UInt160, str]) -> None:
        """
        Helper to fetch all public methods of a smart contract, print their signatures in Python syntax as
        to help determine the right native argument types.

        Note:
            Only native contracts can be queried by their name. Name is case-insensitive.
        """
        state = await self.get_contract_state(contract_hash_or_name)

        print(f"Contract: {state.manifest.name}")
        print((10 + len(state.manifest.name)) * "-")
        for method in state.manifest.abi.methods:
            params = map(lambda p: f", {p.name}: {self._contract_param_to_native(p.type)}", method.parameters)
            params = ''.join(params)  # type: ignore

            # return types are not included because ABI types like ARRAY cannot be properly translated e.g. the
            # following functions both have ARRAY as return type in the ABI but their actual response is very different
            # 1. NeoToken.GetNextBlockValidators returns a list of serialized ECPoints
            # 2. ManagementContract.getContract a serialized ContractState (not even a list)
            print(f"def {method.name}(self{params})")
        print(" ")
        print("ContractParam = Union[bool, int, str, bytes, UInt160, UInt256, ECPoint, list[ContractParam], "
              "Dict[ContractParam, ContractParam]")

    def _contract_param_to_native(self, p):
        if p == contracts.ContractParameterType.BOOLEAN:
            return "bool"
        elif p == contracts.ContractParameterType.INTEGER:
            return "int"
        elif p == contracts.ContractParameterType.STRING:
            return "str"
        elif p == contracts.ContractParameterType.BYTEARRAY:
            return "bytes"
        elif p == contracts.ContractParameterType.HASH160:
            return "UInt160"
        elif p == contracts.ContractParameterType.HASH256:
            return "UInt256"
        elif p == contracts.ContractParameterType.PUBLICKEY:
            return "ECPoint"
        elif p == contracts.ContractParameterType.ARRAY:
            return "list"
        elif p == contracts.ContractParameterType.MAP:
            return "dict"
        elif p == contracts.ContractParameterType.VOID:
            return "None"
        elif p == contracts.ContractParameterType.ANY:
            return "ContractParam"
        else:
            return f"Unknown({str(p)}"



from __future__ import annotations
import abc
from enum import IntEnum
from neo3.core import serialization, Size as s


class NodeCapabilityType(IntEnum):
    #: Server has TCP listening capabilities
    TCPSERVER = 0x01
    #: Server has WebSocket listening capabilities
    WSSERVER = 0x02
    #: Server has full chain data
    FULLNODE = 0x10


class NodeCapability(serialization.ISerializable):
    """
    Capability base class.
    """
    def __init__(self, n_type: NodeCapabilityType):
        self.type = n_type

    def __len__(self):
        """ Get the total size in bytes of the object. """
        return s.uint8

    def __eq__(self, other):
        pass

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type)
        self.serialize_without_type(writer)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.type = NodeCapabilityType(reader.read_uint8())
        self.deserialize_without_type(reader)

    @staticmethod
    def deserialize_from(reader: serialization.BinaryReader) -> NodeCapability:
        capability_type = NodeCapabilityType(reader.read_uint8())
        if capability_type in [NodeCapabilityType.TCPSERVER, NodeCapabilityType.WSSERVER]:
            capability = ServerCapability(capability_type)  # type: NodeCapability
        elif capability_type == NodeCapabilityType.FULLNODE:
            capability = FullNodeCapability()

        capability.deserialize_without_type(reader)
        return capability  # a type of NodeCapability or inherited

    @abc.abstractmethod
    def deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        """ Deserialize from a buffer without reading the `type` member. """

    @abc.abstractmethod
    def serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        """ Serialize into a buffer without including the `type` member. """

    @classmethod
    def _serializable_init(cls):
        return cls(NodeCapabilityType.FULLNODE)


class ServerCapability(NodeCapability):
    """
    A capability expressing node support for TCP or Websocket services.
    """
    def __init__(self, n_type: NodeCapabilityType, port: int = 0):
        super(ServerCapability, self).__init__(n_type)
        if n_type not in [NodeCapabilityType.TCPSERVER, NodeCapabilityType.WSSERVER]:
            raise TypeError(f"{n_type} not one of: {NodeCapabilityType.TCPSERVER.name} {NodeCapabilityType.WSSERVER.name}")  # noqa
        self.port = port

    def __len__(self):
        return super(ServerCapability, self).__len__() + s.uint16

    def serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream without serializing the base class `type` property.

        Args:
            writer: instance.
        """
        writer.write_uint16(self.port)

    def deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream without deserializing the base class `type` property.

        Args:
            reader: instance.
        """
        self.port = reader.read_uint16()

    @classmethod
    def _serializable_init(cls):
        return cls(NodeCapabilityType.TCPSERVER, 0)


class FullNodeCapability(NodeCapability):
    """
    A capability expressing the node has full blockchain data and accepts relaying.
    """
    def __init__(self, start_height: int = 0):
        super(FullNodeCapability, self).__init__(NodeCapabilityType.FULLNODE)
        self.start_height = start_height

    def __len__(self):
        return super(FullNodeCapability, self).__len__() + s.uint32

    def serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream without serializing the base class `type` property.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.start_height)

    def deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream without deserializing the base class `type` property.

        Args:
            reader: instance.
        """
        self.start_height = reader.read_uint32()



from __future__ import annotations
from neo3 import vm
from neo3.core import serialization
from typing import NamedTuple, Union, List, cast


class PlaceHolder(NamedTuple):
    type: vm.StackItemType
    count: int  # type: ignore


class BinarySerializer:

    @staticmethod
    def serialize(stack_item: vm.StackItem, max_size: int) -> bytes:
        """
        Serialize a stack item.

        Note: Interop and Pointer stack items are not supported.

        Args:
            stack_item: the stack item to serialize.
            max_size: maximum byte array output size.

        Raises:
            ValueError: when a circular reference exists in a Map, Struct or Array.
            ValueError: if the output exceeds `max_size`.
        """
        unserialized = [stack_item]
        serialized: List[Union[vm.StackItem, PlaceHolder]] = []
        with serialization.BinaryWriter() as writer:
            while len(unserialized) > 0:
                item = unserialized.pop()
                item_type = type(item)
                writer.write_uint8(int(item.get_type()))

                if item_type == vm.NullStackItem:
                    continue
                elif item_type == vm.BooleanStackItem:
                    writer.write_bool(item.to_boolean())
                elif item_type in [vm.IntegerStackItem, vm.ByteStringStackItem, vm.BufferStackItem]:
                    writer.write_var_bytes(item.to_array())
                elif item_type in [vm.ArrayStackItem, vm.StructStackItem]:
                    if any(map(lambda i: id(i) == id(item), serialized)):
                        raise ValueError("Item already exists")
                    serialized.append(item)
                    writer.write_var_int(len(item))  # type: ignore
                    for element in reversed(item):  # type: ignore
                        unserialized.append(element)
                elif item_type == vm.MapStackItem:
                    if any(map(lambda i: id(i) == id(item), serialized)):
                        raise ValueError("Item already exists")
                    serialized.append(item)
                    writer.write_var_int(len(item))  # type: ignore
                    for k, v in reversed(item):  # type: ignore
                        unserialized.append(v)
                        unserialized.append(k)
                else:
                    raise ValueError(f"Cannot serialize {item_type}")

            if len(writer) > max_size:
                raise ValueError("Output length exceeds max size")
            return writer.to_array()

    @staticmethod
    def deserialize(data: bytes,
                    max_size: int,
                    reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        """
        Deserialize data into a stack item.

        Args:
            data: byte array of a serialized stack item.
            max_size: data reading limit for Array, Struct and Map types.
            reference_counter: a valid reference counter instance. Get's passed into reference stack items.
        """
        if len(data) == 0:
            raise ValueError("Nothing to deserialize")

        deserialized: List[Union[vm.StackItem, PlaceHolder]] = []
        to_deserialize = 1
        with serialization.BinaryReader(data) as reader:
            while not to_deserialize == 0:
                to_deserialize -= 1
                item_type = vm.StackItemType(reader.read_byte()[0])
                if item_type == vm.StackItemType.ANY:
                    deserialized.append(vm.NullStackItem())
                elif item_type == vm.StackItemType.BOOLEAN:
                    deserialized.append(vm.BooleanStackItem(reader.read_bool()))
                elif item_type == vm.StackItemType.INTEGER:
                    deserialized.append(
                        vm.IntegerStackItem(vm.BigInteger(
                            reader.read_var_bytes(vm.IntegerStackItem.MAX_SIZE)
                        ))
                    )
                elif item_type == vm.StackItemType.BYTESTRING:
                    deserialized.append(vm.ByteStringStackItem(reader.read_var_bytes(len(data))))
                elif item_type == vm.StackItemType.BUFFER:
                    deserialized.append(vm.BufferStackItem(reader.read_var_bytes(len(data))))
                elif item_type in [vm.StackItemType.ARRAY, vm.StackItemType.STRUCT]:
                    count = reader.read_var_int(max_size)
                    deserialized.append(PlaceHolder(item_type, count))
                    to_deserialize += count
                elif item_type == vm.StackItemType.MAP:
                    count = reader.read_var_int(max_size)
                    deserialized.append(PlaceHolder(item_type, count))
                    to_deserialize += count * 2
                else:
                    raise ValueError("Invalid format")

        temp: List[vm.StackItem] = []
        while len(deserialized) > 0:
            item = deserialized.pop()
            if type(item) == PlaceHolder:
                item = cast(PlaceHolder, item)
                if item.type == vm.StackItemType.ARRAY:
                    array = vm.ArrayStackItem(reference_counter)
                    for _ in range(0, item.count):
                        array.append(temp.pop())
                    temp.append(array)
                elif item.type == vm.StackItemType.STRUCT:
                    struct = vm.StructStackItem(reference_counter)
                    for _ in range(0, item.count):
                        struct.append(temp.pop())
                    temp.append(struct)
                elif item.type == vm.StackItemType.MAP:
                    m = vm.MapStackItem(reference_counter)
                    for _ in range(0, item.count):
                        k = temp.pop()
                        k = cast(vm.PrimitiveType, k)
                        v = temp.pop()
                        m[k] = v
                    temp.append(m)
            else:
                item = cast(vm.StackItem, item)
                temp.append(item)
        return temp.pop()



from tests.storage import storagetest
from neo3.storage import implementations


class InMemoryDBBlocksTestCase(storagetest.AbstractBlockStorageTest):
    def db_factory(self):
        return implementations.MemoryDB()


class InMemoryDBContractsTestCase(storagetest.AbstractContractStorageTest):
    def db_factory(self):
        return implementations.MemoryDB()


class InMemoryDBStorageTestCase(storagetest.AbstractStorageStorageTest):
    def db_factory(self):
        return implementations.MemoryDB()


class InMemoryDBTransactionsTestCase(storagetest.AbstractTransactionStorageTest):
    def db_factory(self):
        return implementations.MemoryDB()


from __future__ import annotations
import abc
from neo3 import storage
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from neo3.network import payloads


class Snapshot:
    def __init__(self):
        self._block_cache: storage.CachedBlockAccess = None
        self._contract_cache: storage.CachedContractAccess = None
        self._storage_cache: storage.CachedStorageAccess = None
        self._tx_cache: storage.CachedTXAccess = None
        self._block_height_cache: storage.AttributeCache = None
        self.persisting_block: payloads.Block = None

    @property
    def blocks(self):
        return self._block_cache

    @property
    def contracts(self):
        return self._contract_cache

    @property
    def storages(self):
        return self._storage_cache

    @property
    def transactions(self):
        return self._tx_cache

    @property
    def best_block_height(self) -> int:
        try:
            return self._block_height_cache.get()
        except KeyError:
            return -1

    @best_block_height.setter
    def best_block_height(self, value) -> None:
        self._block_height_cache.put(value)

    def commit(self):
        """
        Persist changes to the actual backend.
        """
        self._block_cache.commit()
        self._contract_cache.commit()
        self._storage_cache.commit()
        self._tx_cache.commit()
        self._block_height_cache.commit()

    def clone(self) -> CloneSnapshot:
        return CloneSnapshot(self)

    @abc.abstractmethod
    def __enter__(self):
        """"""

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup resources if necessary."""


class CloneSnapshot(Snapshot):
    def __init__(self, snapshot: Snapshot):
        super(CloneSnapshot, self).__init__()
        self._snapshot = snapshot
        self._block_cache = snapshot.blocks.create_snapshot()
        self._contract_cache = snapshot.contracts.create_snapshot()
        self._storage_cache = snapshot.storages.create_snapshot()
        self._tx_cache = snapshot.transactions.create_snapshot()
        self._block_height_cache = snapshot._block_height_cache.create_snapshot()
        self.persisting_block = snapshot.persisting_block

    def commit(self):
        """
        Persist changes to the parent snapshot.
        """
        super(CloneSnapshot, self).commit()

    def __enter__(self):
        yield self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import json
import binascii
from io import BytesIO
from neo3 import vm
from typing import Union, Any, TypeVar, Iterable, cast


class _JSONDecodeError(json.JSONDecodeError):
    """
    Custom version that doesn't require the document and position arguments

    Can still be caught with json.JSONDecodeError
    """
    def __init__(self, msg):
        ValueError.__init__(self, msg)

    def __reduce__(self):
        return self.__class__, ()  # pragma: no cover


class _NEOVMEncoder(json.JSONEncoder):
    def _encode(self, item):
        t = type(item)
        if t == vm.ArrayStackItem:
            a = []
            for i in item:
                a.append(self._encode(i))
            return a
        elif t == vm.ByteStringStackItem:
            return bytes.fromhex(str(item)).decode()
        elif t == vm.IntegerStackItem:
            i = item.to_biginteger()
            if i > JSONSerializer.MAX_SAFE_INTEGER or i < JSONSerializer.MIN_SAFE_INTEGER:
                return str(i)
            return int(i)
        elif t == vm.BooleanStackItem:
            return item.to_boolean()
        elif t == vm.NullStackItem:
            return None
        elif t == vm.MapStackItem:
            d = {}
            for k, v in item:
                if type(k) == vm.ByteStringStackItem:
                    d.update({k.to_array().decode(): self._encode(v)})
                else:
                    d.update({self._encode(k): self._encode(v)})
            return d
        else:
            raise ValueError

    def default(self, obj):
        try:
            return self._encode(obj)
        except ValueError:
            return json.JSONEncoder.default(self, obj)


class NEOJson:
    """
    JSON wrapper which follows NEOs JSON configuration.
    """
    @staticmethod
    def _float_hook(val):
        # C# strips .0 from floats
        val = float(val)
        if val.is_integer():
            return int(val)
        return val

    @staticmethod
    def _check_for_duplicate_keys(objects):
        # The C# implementation does not allow repeated key names in objects
        unique = dict()
        for pair in objects:
            if pair[0] in unique:
                raise _JSONDecodeError("Duplicate keys in objects are not allowed")
            else:
                unique.update({pair[0]: pair[1]})
        return unique

    @staticmethod
    def _calc_depth(json_input, max_depth, ctr=0):
        if not isinstance(json_input, Iterable):
            return ctr
        if ctr + 1 > max_depth:
            raise ValueError
        ctr += 1

        t = type(json_input)
        if t == list:
            _max = ctr
            for el in json_input:
                ctr = NEOJson._calc_depth(el, max_depth, ctr)
                _max = max(_max, ctr)
            return _max
        elif t == dict:
            _max = ctr
            for k, v in json_input.items():
                ctr = NEOJson._calc_depth(v, max_depth, ctr)
                _max = max(_max, ctr)
            return _max
        else:
            return ctr

    @staticmethod
    def loads(src: Union[str, bytes], max_depth: int = 100) -> Any:
        """
        Parse input into Python objects.

        This is a thin wrapper around the standard library JSON module with adjustments to match NEOs parsing rules.

        Args:
            src: JSON document
            max_depth: maximum nested object depth

        Raises:
            json.JSONDecodeError:
                - if max_depth is exceeded
                - if repeated keys are found
        """

        # Hooks don't get called when subclassing JSONDecoder and trying to provide the hooks from there.
        # so we'll have to stick with the work around until that's fixed.
        d = json.loads(src, parse_float=NEOJson._float_hook, object_pairs_hook=NEOJson._check_for_duplicate_keys)

        # Unfortunately Python does not provide hooks for JSON arrays, only for JSON objects. This means we cannot limit
        # the max depth while parsing and we have to calculate it once the JSON document is fully processed *shrug*
        # Regardless, we're always protected against an infinite recursion by Python's build in stack recursion limit.
        try:
            NEOJson._calc_depth(d, max_depth)
        except ValueError:
            raise _JSONDecodeError("Maximum depth exceeded")
        return d

    @staticmethod
    def dumps(src: dict) -> str:
        """
        Convert `src` to JSON formatted string.

        This is a thin wrapper around the standard library JSON module with adjustments to match NEOs formatting rules.

        Args:
            src: input data.
        """
        return json.dumps(src, cls=_NEOVMEncoder, separators=(',', ':'))


JObject = Union[dict, bool, None, int, list, str]


class JSONSerializer:
    """
    Interoperability layer JSON serialization support
    """
    MAX_SAFE_INTEGER = pow(2, 53) - 1
    MIN_SAFE_INTEGER = -MAX_SAFE_INTEGER

    @staticmethod
    def deserialize(json_data: JObject, reference_counter: vm.ReferenceCounter = None) -> vm.StackItem:
        """
        Deserialize JSON into a virtual machine stack item
        """
        t = type(json_data)
        if t == dict:
            json_data = cast(dict, json_data)
            if reference_counter is None:
                raise ValueError("Can't deserialize JSON object without reference counter")
            map_item = vm.MapStackItem(reference_counter)
            for k, v in json_data.items():
                key = vm.ByteStringStackItem(k)
                value = JSONSerializer.deserialize(v, reference_counter)
                map_item[key] = value
            return map_item
        elif t == list:
            if reference_counter is None:
                raise ValueError("Can't deserialize JSON array without reference counter")
            array_item = vm.ArrayStackItem(reference_counter)
            json_data = cast(list, json_data)
            elements = [JSONSerializer.deserialize(e, reference_counter) for e in json_data]
            array_item.append(elements)
            return array_item
        elif json_data is None:
            return vm.NullStackItem()
        elif t == str:
            if json_data == "null":
                return vm.NullStackItem()
            return vm.ByteStringStackItem(json_data)  # type: ignore
        elif t == int:
            return vm.IntegerStackItem(json_data)  # type: ignore
        elif t == bool:
            json_data = cast(bool, json_data)
            return vm.BooleanStackItem(json_data)
        else:
            # should never happen or somebody ignored the type checker output
            raise ValueError()

    @staticmethod
    def serialize(item: vm.StackItem, max_size: int) -> str:
        """
        Serialize a stack item to a JSON formatted string.

        Args:
            item: a stack item instance.
            max_size: maximum length of serialized data.
        """
        s = json.dumps(item, cls=_NEOVMEncoder, separators=(',', ':'))
        if len(s) > max_size:
            raise ValueError
        return s



import unittest
import binascii
from neo3 import settings, contracts, blockchain
from neo3.core import cryptography, to_script_hash


class TestNativeContract(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        settings.network.standby_committee = ['02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765']
        settings.network.validators_count = 1
        cls.validator_public_key = cryptography.ECPoint.deserialize_from_bytes(
            binascii.unhexlify(settings.network.standby_committee[0])
        )
        cls.validator_account = to_script_hash(
            contracts.Contract.create_multisig_redeemscript(1, [cls.validator_public_key]))

        blockchain.Blockchain()

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_requesting_contract_by_name(self):
        self.assertIsNone(contracts.NativeContract.get_contract_by_name("bogus_contract"))
        self.assertIsInstance(contracts.NativeContract.get_contract_by_name("PolicyContract"), contracts.PolicyContract)

    def test_various(self):
        native = contracts.NativeContract()
        known_contracts = native.registered_contracts
        self.assertIn(contracts.GasToken(), known_contracts)
        self.assertIn(contracts.NeoToken(), known_contracts)
        self.assertIn(contracts.PolicyContract(), known_contracts)


import string

chars = string.digits + string.ascii_letters
base = len(chars)


def encode_base62(num: int):
    """Encode number in base62, returns a string."""
    if num < 0:
        raise ValueError('cannot encode negative numbers')

    if num == 0:
        return chars[0]

    digits = []
    while num:
        rem = num % base
        num = num // base
        digits.append(chars[rem])
    return ''.join(reversed(digits))



import asynctest
import asyncio
from neo3.network import convenience, node
from neo3 import network_logger, blockchain
from neo3.core import msgrouter
from datetime import datetime


class SyncManagerUtilitiesTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.syncmgr = convenience.SyncManager()

    def setUp(self) -> None:
        self.syncmgr._reset_for_test()

    def test_get_best_stored_block_height_with_empty_cache(self):
        # should take the ledger height as highest value
        with asynctest.patch.object(self.syncmgr, 'ledger') as ledger_mock:
            ledger_mock.height = 1
            best_height = self.syncmgr._get_best_stored_block_height()
            self.assertEqual(1, best_height)

    def test_get_best_stored_block_height_with_items_in_cache(self):
        # should take the cache height as highest value
        block2 = asynctest.MagicMock()
        block2.index = 2
        block3 = asynctest.MagicMock()
        block3.index = 3

        # intentionally putting blocks out of order, so also validate sorting
        self.syncmgr.block_cache = [block3, block2]
        with asynctest.patch.object(self.syncmgr, 'ledger') as ledger_mock:
            ledger_mock.height = 1
            best_height = self.syncmgr._get_best_stored_block_height()
            self.assertEqual(3, best_height)

    def test_add_block_flight_info(self):
        # first test creation of a new item
        self.syncmgr.block_requests = {}
        target_height = 1
        self.syncmgr._add_block_flight_info(nodeid=123, height=target_height)
        self.assertIn(target_height, self.syncmgr.block_requests)
        self.assertIsInstance(self.syncmgr.block_requests[target_height], convenience.RequestInfo)
        flights = self.syncmgr.block_requests[target_height].flights
        self.assertEqual(1, len(flights))

        # now test updating with a new flight
        self.syncmgr._add_block_flight_info(nodeid=456, height=target_height)
        flights = self.syncmgr.block_requests[target_height].flights
        self.assertEqual(2, len(flights))

    def test_is_in_blockcache(self):
        block = asynctest.MagicMock()
        block.index = 2
        self.syncmgr.block_cache = [block]

        found = self.syncmgr._is_in_blockcache(block_height=3)
        self.assertFalse(found)
        found = self.syncmgr._is_in_blockcache(block_height=2)
        self.assertTrue(found)


class SyncManagerSyncBlocksTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.syncmgr = convenience.SyncManager()
        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.syncmgr._reset_for_test()
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()

    async def test_sync_blocks_outstanding_requests(self):
        self.syncmgr.block_requests.update({'1': object()})
        self.assertEqual(-1, await self.syncmgr._sync_blocks())

    async def test_sync_blocks_cache_full(self):
        self.syncmgr.block_cache = self.syncmgr.BLOCK_MAX_CACHE_SIZE * [None]
        self.assertEqual(-2, await self.syncmgr._sync_blocks())

    async def test_sync_blocks_no_nodes(self):
        self.assertEqual(-3, await self.syncmgr._sync_blocks())

    # async def test_sync_blocks_no_nodes_with_required_height(self):
    #     mock_node = asynctest.MagicMock()
    #     mock_node.best_height = 1
    #     self.nodemgr.nodes = [mock_node]
    #     with asynctest.patch.object(self.nodemgr, 'get_node_with_height', return_value=None):
    #         self.assertEqual(-4, await self.syncmgr._sync_blocks())

    async def test_sync_blocks_get_1_block(self):
        # scenario
        # - cache has 2 spot available
        # - the connected nodes best_height can only fill 1 of the spots
        # expected result: 1 flight added

        self.syncmgr.block_cache = (self.syncmgr.BLOCK_MAX_CACHE_SIZE - 2) * [None]
        mock_node = node.NeoNode(protocol=object())
        mock_node.best_height = 2
        mock_node.request_block_data = asynctest.CoroutineMock()
        self.nodemgr.nodes = [mock_node]

        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with asynctest.patch.object(self.syncmgr, '_get_best_stored_block_height', return_value=1):
                with asynctest.patch.object(self.syncmgr, '_add_block_flight_info') as mocked_add_flight_info:
                    await self.syncmgr._sync_blocks()
        mocked_add_flight_info.assert_called_once()
        self.assertIn("Asking for blocks 2 - 2", log_context.output[0])


class SyncManagerCheckTimeoutTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.syncmgr = convenience.SyncManager()
        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.syncmgr._reset_for_test()
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()

    async def test_no_outstanding_request(self):
        self.assertEqual(-1, await self.syncmgr._check_timeout())

    async def test_no_flights_timedout(self):
        target_height = 1
        request_info = convenience.RequestInfo(target_height)
        request_info.add_new_flight(convenience.FlightInfo(node_id=123, height=target_height))
        self.syncmgr.block_requests[1] = request_info

        # a recently recreated flight should not have timed out
        self.assertEqual(-2, await self.syncmgr._check_timeout())

    async def test_flights_timed_out(self):
        # scenario: have 2 outstanding flights that timed out
        # - 1 flight for a request that is still is not completed
        # - 1 flight has been made obsolete by a secondary for the same request_info and is still in cache waiting to be processed

        # construct flight 1 - request not yet satisfied
        target_height = 2
        node1_id = 123
        request_info = convenience.RequestInfo(target_height)
        request_info.mark_failed_node = asynctest.MagicMock()
        flight_info = convenience.FlightInfo(node_id=node1_id, height=target_height)
        # reduce start time to enforce exceeding timeout treshold
        flight_info.start_time -= self.syncmgr.BLOCK_REQUEST_TIMEOUT + 1
        request_info.add_new_flight(flight_info)
        self.syncmgr.block_requests[target_height] = request_info

        # construct flight 2 - request already satisfied
        target_height2 = 1
        node2_id = 456
        request_info2 = convenience.RequestInfo(target_height2)
        flight_info2 = convenience.FlightInfo(node_id=node2_id, height=target_height2)
        flight_info2.start_time -= self.syncmgr.BLOCK_REQUEST_TIMEOUT + 1
        request_info2.add_new_flight(flight_info2)
        self.syncmgr.block_requests[target_height2] = request_info2

        # we patch '_get_best_stored_block_height' to return `target_height2` as a way of saying;
        # either the chain or cache already has the data for this height
        with asynctest.patch.object(self.syncmgr, '_get_best_stored_block_height', return_value=target_height2):
            with asynctest.patch.object(self.nodemgr, 'increase_node_timeout_count') as nodemgr_increase_timeout_count:
                result = await self.syncmgr._check_timeout()

        request_info.mark_failed_node.assert_called_with(node1_id)
        # both nodes had a flight that timed out
        nodemgr_increase_timeout_count.assert_has_calls([asynctest.mock.call(node1_id),
                                                         asynctest.mock.call(node2_id)],
                                                        any_order=True)
        # the first time we call it we no longer have any connected nodes, so we can't request from anyone anymore
        self.assertEqual(-3, result)

        # now we "connect" a new node
        mock_node = node.NeoNode(protocol=object())
        mock_node.best_height = 10
        mock_node_id = 789
        mock_node.nodeid = mock_node_id
        mock_node.request_block_data = asynctest.CoroutineMock()
        self.nodemgr.nodes = [mock_node]
        # and try again
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with asynctest.patch.object(self.syncmgr, '_get_best_stored_block_height', return_value=target_height2):
                with asynctest.patch.object(self.nodemgr, 'increase_node_timeout_count'):
                    await self.syncmgr._check_timeout()

        # and validate that a new data request is sent
        self.assertIn("Block timeout for blocks 2 - 2", log_context.output[0])
        mock_node.request_block_data.assert_awaited_once_with(count=1, index_start=2)
        # and also a new flight was added for the new node
        flight = request_info.most_recent_flight()
        self.assertEqual(mock_node_id, flight.node_id)

        # just for coverage
        flight.reset_start_time()
        self.assertTrue(datetime.utcnow().timestamp() - flight.start_time < 0.1)


class SyncManagerVarious(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.syncmgr = convenience.SyncManager()
        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.syncmgr._reset_for_test()
        self.nodemgr._reset_for_test()

    async def test_start_timedout(self):
        with self.assertRaises(Exception) as context:
            with self.assertLogs(network_logger, 'DEBUG') as log_context:
                await self.syncmgr.start(timeout=0.2)
        self.assertIn("Waiting for nodemanager to start", log_context.output[0])
        self.assertIn("Nodemanager failed to start within specified timeout 0.2", str(context.exception))

    async def test_start_shutdown(self):
        self.nodemgr.is_running = True
        # adding an item in the block cache to test that it also runs the `persist_blocks` coroutine
        self.syncmgr.block_cache = [object()]
        with asynctest.patch.object(self.syncmgr, 'persist_blocks', return_value=asynctest.CoroutineMock()):
            with self.assertLogs(network_logger, 'DEBUG') as log_context:
                await self.syncmgr.start(timeout=2)

                self.assertIn("Starting services", log_context.output[1])
                self.assertIsNotNone(self.syncmgr._service_task)

                await asyncio.sleep(0.1)

                self.syncmgr.persist_blocks.assert_awaited_once()

                await self.syncmgr.shutdown()

                for t in self.syncmgr._tasks:
                    self.assertTrue(t.done())

    def test_on_block_received(self):
        # first test receiving a block we have no outstanding request for
        fake_block = asynctest.MagicMock()
        fake_block.index = 1
        fake_block.__len__.return_value = 50
        self.assertEqual(-1, self.syncmgr.on_block_received(from_nodeid=123, block=fake_block))

        # next test receiving a block that we DO have an outstanding request for, but not from the node that is now
        # delivering the block
        request_info = convenience.RequestInfo(height=1)
        request_info.add_new_flight(convenience.FlightInfo(node_id=456, height=1))
        self.syncmgr.block_requests[1] = request_info
        self.assertEqual(-2, self.syncmgr.on_block_received(from_nodeid=123, block=fake_block))
        self.assertEqual(request_info, self.syncmgr.block_requests.get(1, None))

        # next test a valid scenario (outstanding request and receiving a block from the right node)
        mocked_node = node.NeoNode(object())
        mocked_node.nodeweight.append_new_speed = asynctest.MagicMock()
        mocked_node.nodeid = 456
        self.nodemgr.nodes = [mocked_node]
        self.syncmgr.on_block_received(from_nodeid=456, block=fake_block)
        mocked_node.nodeweight.append_new_speed.assert_called_once()
        self.assertIn(fake_block, self.syncmgr.block_cache)
        self.assertEqual(1, len(self.syncmgr.block_cache))

        # and finally try again for the same block and ensure it was not added again to the cache
        self.syncmgr.on_block_received(from_nodeid=456, block=fake_block)
        self.assertEqual(1, len(self.syncmgr.block_cache))

    async def test_persist_blocks(self):

        fake_block1 = asynctest.MagicMock()
        fake_block1.index = 1
        fake_block2 = asynctest.MagicMock()
        fake_block2.index = 2

        # inserting out of order on purpose to also validate sorting
        self.syncmgr.block_cache = [fake_block2, fake_block1]

        mocked_result = asynctest.Mock()
        with asynctest.patch.object(self.syncmgr.ledger, 'persist', side_effect=mocked_result):
            await self.syncmgr.persist_blocks()

        # test that we persisted the blocks in order
        mocked_result.assert_has_calls([asynctest.mock.call(fake_block1),
                                        asynctest.mock.call(fake_block2)],
                                       any_order=False)

        # now test with an exception happening during persist
        self.syncmgr.block_cache = [fake_block2, fake_block1]
        with self.assertLogs(network_logger, 'DEBUG') as log_context:
            with asynctest.patch.object(self.syncmgr.ledger, 'persist', side_effect=Exception):
                await self.syncmgr.persist_blocks()
        self.assertIn("Unexpected exception happened while processing the block cache", log_context.output[0])
        self.assertFalse(self.syncmgr._is_persisting_blocks)

        # test that we don't persist if we're starting to shutdown
        self.syncmgr.shutting_down = True
        mocked_result = asynctest.Mock()
        with asynctest.patch.object(self.syncmgr.ledger, 'persist', side_effect=mocked_result):
            await self.syncmgr.persist_blocks()
        mocked_result.assert_not_called()


class SyncManagerTimedTestCase(asynctest.ClockedTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.syncmgr = convenience.SyncManager()

    def setUp(self) -> None:
        self.syncmgr._reset_for_test()

    async def test_utility_run_in_loop(self):
        counter = 0

        async def coro():
            nonlocal counter
            counter += 1

        self.syncmgr._run_in_loop(coro, 1)
        # trigger the task by ticking the loop once
        await self.advance(0)
        self.assertEqual(1, counter)
        await self.advance(0.9)
        self.assertEqual(1, counter)
        await self.advance(0.1)
        self.assertEqual(2, counter)
        self.syncmgr._tasks[0].cancel()



from __future__ import annotations
from neo3 import contracts, settings
from neo3.core import cryptography
from neo3.contracts.interop import register
from typing import List


CHECKSIG_PRICE = 1 << 15


@register("System.Crypto.CheckSig", CHECKSIG_PRICE, contracts.CallFlags.NONE)
def verify_with_ECDSA_Secp256r1(engine: contracts.ApplicationEngine,
                                public_key: bytes,
                                signature: bytes) -> bool:
    return cryptography.verify_signature(engine.script_container.get_hash_data(settings.network.magic),
                                         signature,
                                         public_key,
                                         cryptography.ECCCurve.SECP256R1)


@register("System.Crypto.CheckMultisig", 0, contracts.CallFlags.NONE)
def check_multisig_with_ECDSA_Secp256r1(engine: contracts.ApplicationEngine,
                                        public_keys: List[bytes],
                                        signatures: List[bytes]) -> bool:
    len_pub_keys = len(public_keys)
    len_sigs = len(signatures)
    if len_sigs == 0:
        raise ValueError("No signatures supplied")
    if len_pub_keys == 0:
        raise ValueError("No public keys supplied")
    if len_sigs > len_pub_keys:
        raise ValueError(f"Verification requires {len_sigs} public keys, got only {len_pub_keys}")

    message = engine.script_container.get_hash_data(settings.network.magic)

    engine.add_gas(len_pub_keys * CHECKSIG_PRICE * engine.exec_fee_factor)

    i = 0
    j = 0
    try:
        while i < len_sigs and j < len_pub_keys:
            if cryptography.verify_signature(message, signatures[i], public_keys[j], cryptography.ECCCurve.SECP256R1):
                i += 1
            j += 1

            if len_sigs - i > len_pub_keys - j:
                return False
    except cryptography.ECCException as e:
        return False
    return True



from __future__ import annotations
from neo3.core import serialization, utils, Size as s, cryptography as crypto


class FilterAddPayload(serialization.ISerializable):
    def __init__(self, data: bytes):
        """
        Create payload.

        Args:
            data: the data to add to the configured bloomfilter.

        """
        self.data = data

    def __len__(self):
        return utils.get_var_size(self.data)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_bytes(self.data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.data = reader.read_var_bytes(520)

    @classmethod
    def _serializable_init(cls):
        return cls(b'')


class FilterLoadPayload(serialization.ISerializable):
    def __init__(self, filter: crypto.BloomFilter):
        """
        Create payload.

        Args:
            filter: bloom filter to load
        """
        self.filter = filter.get_bits()
        self.K = filter.K
        self.tweak = filter.tweak

    def __len__(self):
        return utils.get_var_size(self.filter) + s.uint8 + s.uint32

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_bytes(self.filter)
        writer.write_uint8(self.K)
        writer.write_uint32(self.tweak)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.filter = reader.read_var_bytes(max=36000)
        self.K = reader.read_uint8()
        if self.K > 50:
            raise ValueError("Deserialization error - K exceeds limit of 50")
        self.tweak = reader.read_uint32()

    @classmethod
    def _serializable_init(cls):
        bf = crypto.BloomFilter(8, 2, 345)
        return cls(bf)



import hashlib
from typing import List
from neo3.network import payloads
from neo3.core import types, serialization, to_script_hash
from neo3 import vm, contracts, blockchain, storage


def contract_hash(sender: types.UInt160, checksum: int, name: str) -> types.UInt160:
    sb = vm.ScriptBuilder()
    sb.emit(vm.OpCode.ABORT)
    sb.emit_push(sender.to_array())
    sb.emit_push(checksum)
    sb.emit_push(name)
    return to_script_hash(sb.to_array())


def syscall_name_to_int(name: str) -> int:
    return int.from_bytes(hashlib.sha256(name.encode()).digest()[:4], 'little', signed=False)


def test_engine(has_container=False, has_snapshot=False, default_script=True, call_flags=contracts.CallFlags.ALL):
    tx = payloads.Transaction._serializable_init()

    # this little hack basically nullifies the singleton behaviour and ensures we create
    # a new instance every time we call it. This in turn gives us a clean backend/snapshot
    blockchain.Blockchain.__it__ = None

    snapshot = blockchain.Blockchain(store_genesis_block=True).currentSnapshot
    if has_container and has_snapshot:
        engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, tx, snapshot, 0, test_mode=True)
    elif has_container:
        engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, tx, None, 0, test_mode=True)
    elif has_snapshot:
        engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, None, snapshot, 0, test_mode=True)
    else:
        engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, None, None, 0, test_mode=True)

    if default_script:
        engine.load_script_with_callflags(vm.Script(b'\x40'), call_flags)  # OpCode::RET
    return engine


def test_tx(with_block_height=1, signers: List[types.UInt160]=None) -> payloads.Transaction:
    if signers is None:
        new_signers = [payloads.Signer(types.UInt160.from_string("f782c7fbb2eef6afe629b96c0d53fb525eda64ce"), payloads.WitnessScope.GLOBAL)]
    else:
        new_signers = list(map(lambda v: payloads.Signer(v, payloads.WitnessScope.GLOBAL), signers))

    tx = payloads.Transaction(version=0,
                              nonce=123,
                              system_fee=456,
                              network_fee=789,
                              valid_until_block=with_block_height + 1,
                              attributes=[],
                              signers=new_signers,
                              script=b'\x01',
                              witnesses=[])
    tx.block_height = with_block_height
    return tx


def test_block(with_index=1) -> payloads.Block:
    tx = test_tx(with_index)
    header1 = payloads.Header(
        version=0,
        prev_hash=types.UInt256.from_string("f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
        timestamp=123,
        nonce=0,
        index=with_index,
        primary_index=0,
        next_consensus=types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654"),
        witness=payloads.Witness(invocation_script=b'', verification_script=b'\x55')
    )
    block1 = payloads.Block(header1, transactions=[tx])
    block1.rebuild_merkle_root()
    return block1


class TestIVerifiable(payloads.IVerifiable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.script_hashes = [types.UInt160.zero()]

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        pass

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        pass

    def __len__(self):
        pass

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        pass

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        pass

    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        return self.script_hashes



import asynctest
import asyncio
import logging
from neo3.network import node, message, payloads, capabilities

class NetworkTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        loop = asyncio.get_event_loop()
        loop.set_debug(True)
        def my_exc_handler(loop, context):
            print(f"woahhh error! {context}")
            loop.default_exception_handler(context)
        loop.set_exception_handler(my_exc_handler)

        stdio_handler = logging.StreamHandler()
        stdio_handler.setLevel(logging.DEBUG)
        stdio_handler.setFormatter(logging.Formatter("%(levelname)s - %(module)s:%(lineno)s %(message)s"))
        async_logger = logging.getLogger('asyncio')
        async_logger.addHandler(stdio_handler)
        async_logger.setLevel(logging.DEBUG)

        network_logger = logging.getLogger('neo3.network')
        network_logger.addHandler(stdio_handler)
        network_logger.setLevel(logging.DEBUG)

    @asynctest.SkipTest
    async def test_sending_addresses(self):
        n = await node.NeoNode.connect_to('127.0.0.1', 40333)
        await n.send_address_list([payloads.NetworkAddress(host='127.0.0.1', timestamp=0,
                                capabilities=[capabilities.FullNodeCapability(start_height=123)])])
        await asyncio.sleep(100)

    @asynctest.SkipTest
    async def test_basic_setup(self):
        n = await node.NeoNode.connect_to('127.0.0.1', 40333)
        m = message.Message(message.MessageType.GETFULLBLOCKS, payloads.GetFullBlocksPayload(index_start=1, count=10))
        if n:
            await n.send_message(m)


            # await n.request_address_list()
            await asyncio.sleep(500)
            self.assertIsInstance(n, node.NeoNode)
            # fake nodemanager
            # - add_connected_node
            # - remove_connected_node
            # - quality_check_result
            # - relay_cache.try_get(header)  (can we create relay cache somewhere else?)
            # - sending address list (requires accessing known nodes)




from __future__ import annotations
import hashlib
import typing
from .callflags import CallFlags
from .contracttypes import (TriggerType)
from .descriptor import (ContractPermissionDescriptor)
from .manifest import (ContractGroup,
                       ContractManifest,
                       ContractPermission,
                       WildcardContainer)

from .abi import (ContractParameterType,
                  ContractMethodDescriptor,
                  ContractEventDescriptor,
                  ContractParameterDefinition,
                  ContractABI)
from .nef import (NEF, MethodToken)
from .contract import (Contract, ContractState)
from .binaryserializer import BinarySerializer
from .jsonserializer import (NEOJson, JSONSerializer)
from .findoptions import FindOptions
from .applicationengine import ApplicationEngine
from .native import (NativeContract,
                     PolicyContract,
                     NeoToken,
                     GasToken,
                     OracleContract,
                     DesignationContract,
                     DesignateRole,
                     ManagementContract,
                     LedgerContract,
                     CryptoContract,
                     StdLibContract)


def syscall_name_to_int(name: str) -> int:
    return int.from_bytes(hashlib.sha256(name.encode()).digest()[:4], 'little', signed=False)


def validate_type(obj: object, type_: typing.Type):
    if type(obj) != type_:
        raise ValueError(f"Expected type '{type_}' , got '{type(obj)}' instead")
    return obj


__all__ = ['ContractParameterType',
           'TriggerType',
           'ContractMethodDescriptor',
           'ContractEventDescriptor',
           'ContractParameterDefinition',
           'Contract',
           'ContractState',
           'BinarySerializer',
           'NEOJson',
           'JSONSerializer',
           'NativeContract',
           'CallFlags',
           'PolicyContract',
           'ApplicationEngine',
           'syscall_name_to_int',
           'DesignationContract',
           'DesignateRole',
           'LedgerContract',
           'NEF',
           'MethodToken',
           'FindOptions',
           'CryptoContract',
           'StdLibContract']
